"""
Access data provided by the `United States Army Corps of Engineers`_.

Access data from the USACE `Tulsa District Water Control`_ web site.

.. _United States Army Corps of Engineers: http://www.usace.army.mil/
.. _Tulsa District Water Control: http://www.swt-wc.usace.army.mil/

"""
import datetime

import pandas as pd

from tsgettoolbox.ulmo.usace.swtwc.core import get_station_data

# def get_station_data(station_code, date=None, as_dataframe=False):


def ulmo_df(station_code,
            date=None):
    if date is None:
        date = datetime.datetime.now()
    else:
        date = pd.to_datetime(date)
    alldict = get_station_data(station_code,
                               date=date,
                               as_dataframe=True)
    df = alldict['values']
    df.columns = ['{0}:{1}'.format(i, alldict['variables'][i]['unit'])
                  for i in df.columns]
    df.columns = [i.replace('  ', '_').replace(' ', '_') for i in df.columns]
    df = df.tz_localize(alldict['timezone'])
    return df


if __name__ == '__main__':
    #    import time
    #
    #    r = ulmo_df('blah',
    #                'upperbasin')
    #
    #    print('UB EVERYTHING')
    #    print(r)
    #
    #    time.sleep(20)

    r = ulmo_df('PTTK1',
                date='2017-11-30')

    print('PTTK1 EVERYTHING')
    print(r)



from __future__ import print_function
from __future__ import absolute_import

import logging
import os
from builtins import object
from io import BytesIO

from tsgettoolbox.odo import convert
from tsgettoolbox.odo import odo
from tsgettoolbox.odo import resource

import pandas as pd

import requests

from tstoolbox import tsutils

# unavco

# pd.read_csv('http://web-services.unavco.org:80/met/data/P0005/beta?starttime=2012-05-01T00%3A00%3A00&endtime=2012-05-02T23%3A59&tsFormat=iso8601'


class Unavco(object):
    def __init__(self, url, **query_params):
        try:
            station = query_params.pop('station')
        except KeyError:
            raise KeyError("""
*
*   The station keyword is required.  You have given me
*   {0}.
*
""".format(query_params))
        self.url = '{0}/{1}/beta'.format(url, station)
        if '/met/' in self.url or '/strain/' in self.url:
            self.comment = None
        else:
            self.comment = '#'
        query_params['starttime'] = tsutils.parsedate(
            query_params['starttime']).isoformat()
        query_params['endtime'] = tsutils.parsedate(
            query_params['endtime']).isoformat()
        self.query_params = query_params
        self.query_params['tsFormat'] = 'iso8601'

# Function to make `resource` know about the new unavco type.


@resource.register(r'http://web-services.unavco.org:80/met/data.*',
                   priority=17)
@resource.register(r'http://web-services.unavco.org:80/pore/data/temperature.*',
                   priority=17)
@resource.register(r'http://web-services.unavco.org:80/pore/data/pressure.*',
                   priority=17)
@resource.register(r'http://web-services.unavco.org:80/tilt/data.*',
                   priority=17)
@resource.register(r'http://web-services.unavco.org:80/strain/data/L2.*',
                   priority=17)
def resource_unavco(uri, **kwargs):
    return Unavco(uri, **kwargs)

# Function to convert from Unavco type to pd.DataFrame


@convert.register(pd.DataFrame, Unavco)
def unavco_to_df(data, **kwargs):
    req = requests.get(data.url,
                       params=data.query_params)
    if os.path.exists('debug_tsgettoolbox'):
        logging.warning(req.url)
    df = pd.read_csv(BytesIO(req.content),
                     header=0,
                     index_col=0,
                     parse_dates=[0],
                     comment=data.comment)
    df.columns = ['unavco-{0}'.format(i.
                                      strip().
                                      replace(' ', '_').
                                      replace('(', ':').
                                      replace(')', '').
                                      replace('deg._C', 'degC'))
                  for i in df.columns]
    df.index.name = 'Datetime:UTC'
    return df.tz_localize('UTC')


if __name__ == '__main__':
    r = resource(
        r'http://web-services.unavco.org:80/met/data',
        station='P005',
        starttime='2012-05-01T00:00:00',
        endtime='2012-05-02T23:59:59'
    )

    as_df = odo(r, pd.DataFrame)
    print('Unavco_met')
    print(as_df)

    r = resource(
        r'http://web-services.unavco.org:80/pore/data/temperature',
        station='B078',
        starttime='2012-05-02T00:00:00',
        endtime='2012-05-02T23:59:59'
    )

    as_df = odo(r, pd.DataFrame)
    print('Unavco_pore_temperature')
    print(as_df)

    r = resource(
        r'http://web-services.unavco.org:80/pore/data/pressure',
        station='B078',
        starttime='2012-05-02T00:00:00',
        endtime='2012-05-02T23:59:59'
    )

    as_df = odo(r, pd.DataFrame)
    print('Unavco_pore_pressure')
    print(as_df)

    r = resource(
        r'http://web-services.unavco.org:80/tilt/data',
        station='P693',
        starttime='2012-05-01T00:00:00',
        endtime='2012-05-01T01:00:00'
    )

    as_df = odo(r, pd.DataFrame)
    print('Unavco_tilt')
    print(as_df)

    r = resource(
        r'http://web-services.unavco.org:80/strain/data/L2',
        station='B007',
        starttime='2012-05-01T00:00:00',
        endtime='2012-05-01T23:59:59'
    )

    as_df = odo(r, pd.DataFrame)
    print('Unavco_strain')
    print(as_df)




from tsgettoolbox.odo import odo, resource
import pandas as pd
import mando

try:
    from mando.rst_text_formatter import RSTHelpFormatter as HelpFormatter
except ImportError:
    from argparse import RawTextHelpFormatter as HelpFormatter

from tstoolbox import tsutils


@mando.command('unavco', formatter_class=HelpFormatter, doctype='numpy')
def unavco_cli(station,
               database='met',
               starttime=None,
               endtime=None):
    r"""Download data from the Unavco web services.

    Detailed information at:
    http://www.unavco.com/data/web-services/web-services.html

    The "database" option defines different return data.

    met::

       Returns hourly meteorological data for the specified station and
       timeframe.  Pressure, temperature, humidity, wind speed and
       direction are provided as averages of all the samples taken per
       hour. Rain and hail are totals for the hour. The sample count is
       the number of samples taken per hour.

       Stations are configured to sample either at 1 minute or 5 minute
       intervals and the user can determine which by looking at the
       sample counts over several hours to see if they approach 12 or
       60.

       Returns: sample timestamp, pressure (mbar), temperature (degree
       C), relative humidity(per cent), wind direction(degrees), wind
       speed(m/s), rain(0.1mm), hail(hits), sample count

    pore_temperature::

       Pore pressure and temperature readings are collected by pore
       pressure sensors co-located with borehole strainmeters. Tilt data
       is collected by shallow borehole tiltmeters co-located with
       borehole strainmeters and seismic stations.

       Get pore temperature for the specified stations and time range.

       Returns: sample time, temperature(degree C)

    pore_pressure::

       Pore pressure and temperature readings are collected by pore
       pressure sensors co-located with borehole strainmeters. Tilt data
       is collected by shallow borehole tiltmeters co-located with
       borehole strainmeters and seismic stations.

       Get pore pressure for the specified stations and time range.

       Returns: sample time, pressure (hPa)

    tilt::

       Get tilt data for specified stations and time range.

       Returns: DateTime, X-axis tilt (microRadians), Y-axis tilt
       (microRadians), Temperature (degree C), Voltage(v)

    strain::

       Geodetic strain data is collected by four-component deep borehole
       tensor strainmeters that record transient deformation signals
       yielding information about the physical properties of the
       surrounding rock.

       Borehole strainmeters measure very small changes in the dimension
       of a borehole at depths ranging from 100m to 250m. The Plate
       Boundary Observatory uses a instrument developed and constructed
       by GTSM Technologies which measure the change in borehole
       diameter along three azimuths separated by 120 degrees
       perpendicular to the borehole.

       Get borehole strain data for the borehole strainmeter station
       identified.  This data is low rate, 5 minute sample, level
       2 uncorrected and corrected strain data. Corrected values are the
       uncorrected strain minus the effects of the tidal signal and
       barometric pressure.

       Returns: DateTime, Gauge 0 uncorrected microstrain, Gauge
       1 corrected microstrain, Gauge 1 uncorrected microstrain, Gauge
       1 corrected microstrain, Gauge 2 microstrain, Gauge 2 corrected
       microstrain, Gauge 3 microstrain, Gauge 3 corrected microstrain,
       Eee+Enn uncorrected microstrain, Eee+Enn corrected microstrain,
       Eee-Enn uncorrected microstrain, Eee-Enn corrected microstrain,
       2Ene uncorrected microstrain, 2Ene corrected microstrain.

    Parameters
    ----------
    station
        Unavco station identifier
    database : str
        Database to pull from.  One of 'met', 'pore_temperature',
        'pore_pressure', 'tilt', 'strain'.  The default is 'met'.
    starttime
        Start date in ISO8601 format.
    endtime
        End date in ISO8601 format.
    """
    tsutils._printiso(unavco(station,
                             database=database,
                             starttime=starttime,
                             endtime=endtime))


def unavco(station,
           database='met',
           starttime=None,
           endtime=None):
    r"""Download data from the Unavco web services."""
    from tsgettoolbox.services import unavco as placeholder
    map_db_to_url = {
        'met': r'http://web-services.unavco.org:80/met/data',
        'pore_temperaure': r'http://web-services.unavco.org:80'
                           '/pore/data/temperature',
        'pore_pressure': r'http://web-services.unavco.org:80'
                         '/pore/data/pressure',
        'tilt': r'http://web-services.unavco.org:80/tilt/data',
        'strain': r'http://web-services.unavco.org:80/strain/data/L2',
    }
    r = resource(
        map_db_to_url[database],
        station=station,
        starttime=starttime,
        endtime=endtime,
    )
    return odo(r, pd.DataFrame)




from __future__ import absolute_import
from __future__ import print_function

import logging
import os
from io import BytesIO

from tsgettoolbox.odo import convert
from tsgettoolbox.odo import odo
from tsgettoolbox.odo import resource

import pandas as pd

import requests

from tstoolbox import tsutils

# USGS

# parameter_cd	parameter_group_nm	parameter_nm	casrn	srsname	parameter_units
pmcodes = pd.read_csv(os.path.join(os.path.dirname(__file__), 'pmcodes.dat'),
                      comment='#',
                      header=0,
                      sep='\t',
                      dtype={0: str},
                      na_values='Dis')
pmcodes.set_index('parameter_cd', inplace=True)

# IV
#
# agency_cd
# site_no
# datetime
# tz_cd
# 30725_00060
# 30725_00060_cd
# 196788_00065
# 196788_00065_cd
#
# DV
#
# agency_cd
# site_no
# datetime
# 68479_00010_00001
# 68479_00010_00001_cd
# 68482_00010_00001
# 68482_00010_00001_cd
#
# STAT
#
# agency_cd
# site_no
# station_nm
# site_tp_cd
# dec_lat_va
# dec_long_va
# coord_acy_cd
# dec_coord_datum_cd
# alt_va
# alt_acy_va
# alt_datum_cd
# huc_cd
#
# GWLEVELS
#
# agency_cd
# site_no
# site_tp_cd
# lev_dt
# lev_tm
# lev_tz_cd
# lev_va
# sl_lev_va
# sl_datum_cd
# lev_status_cd
# lev_agency_cd
# lev_dt_acy_cd
# lev_acy_cd
# lev_src_cd
# lev_meth_cd
# lev_age_cd
#
# STATS
#
# agency_cd
# site_no
# parameter_cd
# ts_id
# loc_web_ds
# month_nu
# day_nu
# begin_yr
# end_yr
# count_nu
# max_va_yr
# max_va
# min_va_yr
# min_va
# mean_va
# p05_va
# p10_va
# p20_va
# p25_va
# p50_va
# p75_va
# p80_va
# p90_va
# p95_va


def _read_rdb(data):
    req = requests.get(data.url,
                       params=data.query_params)
    if os.path.exists('debug_tsgettoolbox'):
        logging.warning(req.url)
    req.raise_for_status()

    if '/iv/' in data.url or '/dv/' in data.url:
        # iv and dv results are stacked, a table for each site.  Have to split
        # the overall req.content into discrete tables for pd.read_csv to work.
        list_of_sublists = []
        n = 0
        a_list = req.content.splitlines()
        for i, elt in enumerate(a_list):
            if i and elt[:9] == b'agency_cd':
                list_of_sublists.append(a_list[n:i])
                n = i
        list_of_sublists.append(a_list[n:])

        ndf = pd.DataFrame()
        for site in list_of_sublists:
            try:
                adf = pd.read_csv(BytesIO(b'\n'.join(site)),
                                  comment='#',
                                  header=[0, 1],
                                  sep='\t',
                                  dtype={'site_no': str},
                                  na_values='Dis')
            except pd.errors.EmptyDataError:
                continue

            adf.columns = [i[0] for i in adf.columns]

            test_cnames = []
            not_ts = []
            for cname in adf.columns:
                words = cname.split('_')
                try:
                    _ = int(words[0])
                    if 'cd' == words[-1]:
                        test_cnames.append(cname)
                    else:
                        test_cnames.append(cname + ':{0}'.format(
                            pmcodes.loc[words[1], 'parameter_units']))
                except ValueError:
                    test_cnames.append(cname)
                    not_ts.append(cname)

            adf.columns = test_cnames
            adf.set_index(not_ts, inplace=True)

            if len(ndf) == 0:
                ndf = adf
            else:
                ndf = ndf.join(adf, how='outer')

        ndf.reset_index(inplace=True)
    else:
        ndf = pd.read_csv(BytesIO(req.content),
                          comment='#',
                          header=[0, 1],
                          sep='\t',
                          dtype={'site_no':str,
                                 'parameter_cd':str,
                                 'ts_id':str},
                          na_values='Dis')
        ndf.columns = [i[0] for i in ndf.columns]
    return ndf


def _make_nice_names(ndf, reverse=False):
    nnames = []
    for col in ndf.columns.values:
        strung = [str(i) for i in col]
        if reverse is True:
            strung = reversed(strung)
        nnames.append('_'.join(strung).strip())
    return nnames


tzmap = {
    'EST': 'America/New_York',
    'EDT': 'America/New_York',
    'CST': 'America/Chicago',
    'CDT': 'America/Chicago',
    'MST': 'America/Denver',
    'MDT': 'America/Denver',
    'PST': 'America/Los_Angeles',
    'PDT': 'America/Los_Angeles',
}


def normalize_tz(row, tz_cd):
    try:
        return row['Datetime'].tz_localize(tzmap[row[tz_cd]])
    except KeyError:
        return row['Datetime']


class USGS_IV_DV_RDB(object):
    def __init__(self, url, **query_params):
        # Need to enforce RDB format
        query_params['format'] = 'rdb'
        query_params['startDT'] = tsutils.parsedate(query_params['startDT'],
                                                    strftime='%Y-%m-%d')
        query_params['endDT'] = tsutils.parsedate(query_params['endDT'],
                                                  strftime='%Y-%m-%d')

        self.includeCodes = True
        if 'includeCodes' in query_params:
            self.includeCodes = query_params.pop('includeCodes')

        self.url = url
        self.query_params = query_params


# Function to make `resource` know about the new USGS iv type.
@resource.register(r'http(s)?://waterservices\.usgs\.gov/nwis/dv/.*',
                   priority=17)
@resource.register(r'http(s)?://waterservices\.usgs\.gov/nwis/iv/.*',
                   priority=17)
def resource_usgs_iv_dv_rdb(uri, **kwargs):
    return USGS_IV_DV_RDB(uri, **kwargs)


@convert.register(pd.DataFrame, USGS_IV_DV_RDB)
def usgs_iv_dv_rdb_to_df(data, **kwargs):
    """Convert from USGS RDB type to pd.DataFrame."""
    ndf = _read_rdb(data)
    ndf['Datetime'] = pd.to_datetime(ndf['datetime'])
    ndf.drop('datetime',
             axis='columns',
             inplace=True)
    if 'tz_cd' in ndf.columns:
        ndf['Datetime'] = ndf.apply(normalize_tz,
                                    args=('tz_cd',),
                                    axis=1)
        ndf.drop('tz_cd',
                 axis='columns',
                 inplace=True)

    ndf.set_index(['agency_cd',
                   'site_no',
                   'Datetime'],
                  inplace=True)
    ndf = ndf.unstack(level=['site_no',
                             'agency_cd'])

    # Sometime in the near future figure out a better way because right now the
    # ndf.unstack above can create a huge dataframe that is mostly NA.
    # Workaround is to trim it down to size in next command.
    ndf.dropna(axis='columns', how='all', inplace=True)

    ndf.columns = _make_nice_names(ndf, reverse=True)

    if data.includeCodes is False:
        ndf.drop([i for i in ndf.columns if i[-3:] == '_cd'],
                 axis='columns',
                 inplace=True)
    return ndf


class USGS_STAT_RDB(object):
    def __init__(self, url, **query_params):

        # set defaults.
        for key, val in [['statYearType', 'calendar'],
                         ['missingData', 'off'],
                         ['statType', 'all'],
                         ['statReportType', 'daily']]:
            try:
                if query_params[key] is None:
                    query_params[key] = val
            except KeyError:
                query_params[key] = val

        # Need to enforce rdb format
        query_params['format'] = 'rdb'

        if query_params['statReportType'] != 'annual':
            query_params['statYearType'] = None
        if query_params['statReportType'] == 'daily':
            query_params['missingData'] = None
        self.url = url
        self.query_params = query_params


@resource.register(r'http(s)?://waterservices\.usgs\.gov/nwis/stat/.*',
                   priority=17)
def resource_usgs_stat_rdb(uri, **kwargs):
    """Make `resource` know about the new USGS stat type."""
    return USGS_STAT_RDB(uri, **kwargs)


@convert.register(pd.DataFrame, USGS_STAT_RDB)
def usgs_stat_rdb_to_df(data, **kwargs):
    """Convert from USGS STAT_RDB type to pd.DataFrame."""
    ndf = _read_rdb(data)
    if data.query_params['statReportType'] == 'daily':
        ndf['Datetime'] = ['{0:02d}-{1:02d}'.format(int(i), int(j))
                           for i, j in zip(ndf['month_nu'], ndf['day_nu'])]
        ndf.drop(['month_nu', 'day_nu'], axis=1, inplace=True)
    elif data.query_params['statReportType'] == 'monthly':
        ndf['Datetime'] = pd.to_datetime(['{0}-{1:02d}'.format(i, int(j))
                                          for i, j in zip(ndf['year_nu'],
                                                          ndf['month_nu'])])
        ndf.drop(['year_nu', 'month_nu'], axis=1, inplace=True)
    else:
        if data.query_params['statYearType'] == 'water':
            ndf['Datetime'] = pd.to_datetime(['{0}-10-01'.format(int(i) - 1)
                                              for i in ndf['year_nu']])
        else:
            ndf['Datetime'] = pd.to_datetime(ndf['year_nu'])
        ndf.drop('year_nu', axis=1, inplace=True)
    ndf.sort_values(['agency_cd',
                     'site_no',
                     'parameter_cd',
                     'ts_id',
                     'Datetime'],
                    inplace=True)
    ndf.set_index(['agency_cd',
                   'site_no',
                   'parameter_cd',
                   'ts_id',
                   'Datetime'],
                  inplace=True)
    ndf = ndf.unstack(level=['agency_cd',
                             'site_no',
                             'parameter_cd',
                             'ts_id'])
    ndf = ndf.reorder_levels([1, 2, 3, 4, 0], axis=1)

    ndf.columns = _make_nice_names(ndf)

    return ndf


statelookup = {
    1: 'AL',  # Alabama
    2: 'AK',  # Alaska
    4: 'AZ',  # Arizona
    5: 'AR',  # Arkansas
    6: 'CA',  # California
    8: 'CO',  # Colorado
    9: 'CT',  # Connecticut
    10: 'DE',  # Delaware
    11: 'DC',  # District of Columbia
    12: 'FL',  # Florida
    13: 'GA',  # Georgia
    15: 'HI',  # Hawaii
    16: 'ID',  # Idaho
    17: 'IL',  # Illinois
    18: 'IN',  # Indiana
    19: 'IA',  # Iowa
    20: 'KS',  # Kansas
    21: 'KY',  # Kentucky
    22: 'LA',  # Louisiana
    23: 'ME',  # Maine
    24: 'MD',  # Maryland
    25: 'MA',  # Massachusetts
    26: 'MI',  # Michigan
    27: 'MN',  # Minnesota
    28: 'MS',  # Mississippi
    29: 'MO',  # Missouri
    30: 'MT',  # Montana
    31: 'NE',  # Nebraska
    32: 'NV',  # Nevada
    33: 'NH',  # New Hampshire
    34: 'NJ',  # New Jersey
    35: 'NM',  # New Mexico
    36: 'NY',  # New York
    37: 'NC',  # North Carolina
    38: 'ND',  # North Dakota
    39: 'OH',  # Ohio
    40: 'OK',  # Oklahoma
    41: 'OR',  # Oregon
    42: 'PA',  # Pennsylvania
    44: 'RI',  # Rhode Island
    45: 'SC',  # South Carolina
    46: 'SD',  # South Dakota
    47: 'TN',  # Tennessee
    48: 'TX',  # Texas
    49: 'UT',  # Utah
    50: 'VT',  # Vermont
    51: 'VA',  # Virginia
    53: 'WA',  # Washington
    54: 'WV',  # West Virginia
    55: 'WI',  # Wisconsin
    56: 'WY',  # Wyoming
}


class USGS_MEASUREMENTS_PEAK_RDB(object):
    def __init__(self, url, **query_params):

        if 'measurements' in url:
            rdb_format = 'rdb_expanded'
        elif 'peak' in url:
            rdb_format = 'rdb'

        # Need to enforce rdb format
        query_params['format'] = rdb_format
        query_params['agency_cd'] = 'USGS'
        query_params['site_no'] = query_params['sites']
        query_params.pop('sites')

        # Get the state code and insert into URL
        r = odo(resource(r'http://waterservices.usgs.gov/nwis/site/',
                         sites=query_params['site_no']),
                pd.DataFrame)
        try:
            url = url.replace('XX',
                              statelookup[int(r.ix[1,
                                                   u'state_cd'])].lower())
        except KeyError:
            raise ValueError("""
*
*   No field measurements available.  Some states don't have any posted
*   to NWIS.
*
""")

        self.url = url
        self.query_params = query_params


# https://nwis.waterdata.usgs.gov/fl/nwis/measurements?site_no=02315500&agency_cd=USGS&format=rdb_expanded
@resource.register(
    r'http(s)?://nwis\.waterdata\.usgs\.gov/../nwis/measurements.*',
    priority=17)
# https://nwis.waterdata.usgs.gov/fl/nwis/peak?site_no=02315500&agency_cd=USGS&format=rdb_expanded
@resource.register(r'http(s)?://nwis\.waterdata\.usgs\.gov/../nwis/peak.*',
                   priority=17)
def resource_usgs_measurements_peak_rdb(uri, **kwargs):
    return USGS_MEASUREMENTS_PEAK_RDB(uri, **kwargs)


@convert.register(pd.DataFrame, USGS_MEASUREMENTS_PEAK_RDB)
def usgs_measurements_peak_rdb_to_df(data, **kwargs):
    """Convert from USGS RDB type to pd.DataFrame."""
    ndf = _read_rdb(data)
    if 'measurements' in data.url:
        dname = 'measurement_dt'
    elif 'peak' in data.url:
        dname = 'peak_dt'
    ndf['Datetime'] = pd.to_datetime(ndf[dname], errors='coerce')
    ndf.set_index(['Datetime'],
                  inplace=True)
    ndf.drop([dname,
              'agency_cd',
              'site_no'],
             axis=1,
             inplace=True)
    return ndf


class USGS_SITE_RDB(object):
    def __init__(self, url, **query_params):

        # Need to enforce rdb format
        query_params['format'] = 'rdb'
        query_params['siteOutput'] = 'expanded'
        query_params['siteStatus'] = 'all'

        self.url = url
        self.query_params = query_params


# https://waterservices.usgs.gov/nwis/site/?format=rdb&sites=01646500&siteOutput=expanded&siteStatus=all
@resource.register(r'http(s)?://waterservices\.usgs\.gov/nwis/site.*',
                   priority=17)
def resource_usgs_site_rdb(uri, **kwargs):
    return USGS_SITE_RDB(uri, **kwargs)


@convert.register(pd.DataFrame, USGS_SITE_RDB)
def usgs_site_rdb_to_df(data, **kwargs):
    """Convert from USGS RDB type to pd.DataFrame."""
    ndf = _read_rdb(data)
    return ndf


class USGS_GWLEVELS_RDB(object):
    def __init__(self, url, **query_params):

        # Need to enforce rdb format
        query_params['format'] = 'rdb'

        self.url = url
        self.query_params = query_params


@resource.register(r'http(s)?://waterservices\.usgs\.gov/nwis/gwlevels/.*',
                   priority=17)
def resource_usgs_gwlevels_rdb(uri, **kwargs):
    """Assign URL to a function.."""
    return USGS_GWLEVELS_RDB(uri, **kwargs)


@convert.register(pd.DataFrame, USGS_GWLEVELS_RDB)
def usgs_gwlevels_rdb(data, **kwargs):
    """Convert from USGS RDB type to pd.DataFrame."""
    ndf = _read_rdb(data)
    # lev_dt    lev_tm  lev_tz_cd
    ndf['Datetime'] = pd.to_datetime(ndf['lev_dt'] +
                                     'T' +
                                     ndf['lev_tm'],
                                     errors='coerce')
    mask = pd.isnull(ndf['Datetime'])
    ndf.loc[mask, 'Datetime'] = pd.to_datetime(ndf.loc[mask, 'lev_dt'],
                                               errors='coerce')
    ndf['Datetime'] = ndf.apply(normalize_tz,
                                args=('lev_tz_cd',),
                                axis=1)
    ndf.set_index(['Datetime'],
                  inplace=True)
    try:
        ndf.index.name = 'Datetime:{0}'.format(ndf.index.tz)
    except AttributeError:
        ndf.index.name = 'Datetime'
    ndf.drop(['lev_dt',
              'lev_tm',
              'lev_tz_cd',
              'site_no'],
             axis=1,
             inplace=True)
    return ndf


if __name__ == '__main__':
    R = resource(
        r'http://waterservices.usgs.gov/nwis/gwlevels/',
        sites='375907091432201',
        startDT='2017-01-01',
        endDT='2017-12-30'
    )
    AS_DF = odo(R, pd.DataFrame)
    print('USGS_GWLEVELS single')
    print(AS_DF)

    R = resource(
        r'http://waterservices.usgs.gov/nwis/gwlevels/',
        hucs='03110201',
        startDT='2017-01-01',
        endDT='2017-12-30'
    )
    AS_DF = odo(R, pd.DataFrame)
    print('USGS_GWLEVELS multiple')
    print(AS_DF)

    R = resource(
        r'http://waterservices.usgs.gov/nwis/iv/',
        sites='02325000',
        startDT='2015-07-01',
        endDT='2015-07-30'
    )
    AS_DF = odo(R, pd.DataFrame)
    print('USGS_IV single')
    print(AS_DF)

    R = resource(
        r'http://waterservices.usgs.gov/nwis/iv/',
        sites='02325000,02239501',
        startDT='2015-07-01',
        endDT='2015-07-30'
    )
    AS_DF = odo(R, pd.DataFrame)
    print('USGS_IV multiple')
    print(AS_DF)

    R = resource(
        r'http://waterservices.usgs.gov/nwis/dv/',
        sites='02325000',
        startDT='2015-07-01',
        endDT='2015-07-30'
    )
    AS_DF = odo(R, pd.DataFrame)
    print('USGS_DV')
    print(AS_DF)

    R = resource(
        r'http://waterservices.usgs.gov/nwis/dv/',
        sites='02325000,02239501',
        startDT='2015-07-01',
        endDT='2015-07-30'
    )
    AS_DF = odo(R, pd.DataFrame)
    print('USGS_DV multiple')
    print(AS_DF)

    R = resource(
        r'http://waterservices.usgs.gov/nwis/stat/',
        sites='02325000',
    )
    AS_DF = odo(R, pd.DataFrame)
    print('USGS_DAILY_STAT single')
    print(AS_DF)

    R = resource(
        r'http://waterservices.usgs.gov/nwis/stat/',
        sites='02325000,02239501',
    )
    AS_DF = odo(R, pd.DataFrame)
    print('USGS_DAILY_STAT multiple')
    print(AS_DF)

    R = resource(
        r'http://waterservices.usgs.gov/nwis/stat/',
        sites='01646500',
        statReportType='monthly',
    )
    AS_DF = odo(R, pd.DataFrame)
    print('USGS_MONTHLY_STAT single')
    print(AS_DF)

    R = resource(
        r'http://waterservices.usgs.gov/nwis/stat/',
        sites='02325000,01646500',
        statReportType='monthly',
    )
    AS_DF = odo(R, pd.DataFrame)
    print('USGS_MONTHLY_STAT multiple')
    print(AS_DF)

    R = resource(
        r'http://waterservices.usgs.gov/nwis/stat/',
        sites='01646500',
        statReportType='annual',
        statYearType='water',
    )
    AS_DF = odo(R, pd.DataFrame)
    print('USGS_ANNUAL_STAT single')
    print(AS_DF)

    R = resource(
        r'http://waterservices.usgs.gov/nwis/stat/',
        sites='01646500,02239501',
        statReportType='annual',
        statYearType='water',
    )
    AS_DF = odo(R, pd.DataFrame)
    print('USGS_ANNUAL_STAT multple')
    print(AS_DF)



#!/usr/bin/env python
r"""
tsgettoolbox command line/library tools to retrieve time series.

This program is a collection of utilities to download data from various
web services.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import warnings

from tsgettoolbox.odo import odo, resource
import pandas as pd
import mando

try:
    from mando.rst_text_formatter import RSTHelpFormatter as HelpFormatter
except ImportError:
    from argparse import RawTextHelpFormatter as HelpFormatter

from tstoolbox import tsutils

warnings.filterwarnings('ignore')


nwis_docstrings = {
    'includeCodes': r"""includeCodes
        [optional, default is False]

        Whether or not to include the metadata/quality code column.  Useful to
        almost halve the size of the pandas DataFrame.  """,
    'sites': r"""sites : str
        [optional, default is None]

        Want to only query one site? Use sites as your major filter, and put
        only one site number in the list.  Sites are comma separated. Sites may
        be prefixed with an optional agency code followed by a colon. If you
        do not know the site numbers you need, you can find relevant sites with
        the NWIS Mapper (http://wdr.water.usgs.gov/nwisgmap/index.html) or on
        the USGS Water Data for the Nation site.
        (http://waterdata.usgs.gov/nwis/)

        Can have from 1 to 100 comma separated site numbers::

            --sites=USGS:01646500
            --sites=01646500,06306300""",
    'stateCd': r"""stateCd : str
        [optional, default is None]

        U.S. postal service (2-digit) state code.  Can have only 1 state code.
        List is available at
        http://www.usps.com/ncsc/lookups/usps_abbreviations.html::

            --stateCd=NY""",
    'huc': r"""huc : str
        [optional, default is None]

        A list of hydrologic unit codes (HUC) or watersheds.  Only 1 major HUC
        can be specified per request.  A major HUC has two digits. Minor HUCs
        must be eight digits in length.  Can have 1 to 10 HUC codes.  List of
        HUCs is available at http://water.usgs.gov/GIS/huc_name.html::

            --huc=01,02070010""",
    'bBox': r"""bBox :
        [optional, default is None]

        A contiguous range of decimal latitude and longitude, starting with the
        west longitude, then the south latitude, then the east longitude, and
        then the north latitude with each value separated by a comma. The
        product of the range of latitude and longitude cannot exceed 25
        degrees. Whole or decimal degrees must be specified, up to six digits
        of precision. Minutes and seconds are not allowed. Remember: western
        longitude (which includes almost all of the United States) is specified
        in negative degrees. Caution: many sites outside the continental US do
        not have latitude and longitude referenced to NAD83 and therefore can
        not be found using these arguments. Certain sites are not associated
        with latitude and longitude due to homeland security concerns and
        cannot be found using this filter.::

            --bBox=-83,36.5,-81,38.5""",
    'countyCd': r"""countyCd :
        [optional, default is None]

        A list of county numbers, in a 5 digit numeric format. The first two
        digits of a county's code are the FIPS State Code.  Can have from 1 to
        20 county codes.  The first 2 digits are the FIPS State Code
        (http://www.itl.nist.gov/fipspubs/fip5-2.htm) and the list of county
        codes are at
        http://help.waterdata.usgs.gov/code/county_query?fmt=html::

            --countyCd=51059,51061""",
    'parameterCd': r"""parameterCd :
        [optional, default is None]

        USGS time-series parameter code.  All parameter codes are numeric and
        5 characters in length.  Parameter codes are used to identify the
        constituent measured and the units of measure.  Popular codes include
        stage (00065), discharge in cubic feet per second (00060) and water
        temperature in degrees Celsius (00010). Can request from 1 to 100
        "parameterCD"s.  Default: returns all regular time-series for the
        requested sites.

        Complete list::

            http://help.waterdata.usgs.gov/codes-and-parameters/parameters::

            --parameterCd=00060       # discharge, cubic feet
                                      # per second
            --parameterCd=00060,00065 # discharge,
                                      # cubic feet per second
                                      # and gage height in
                                      # feet""",
    'siteType': r"""siteType :
        [optional, default is None]

        Restricts sites to those having one or more major and/or minor site
        types.  If you request a major site type (ex: &siteType=ST) you will
        get all sub-site types of the same major type as well (in this case,
        ST-CA, ST-DCH and ST-TS).  Can have from 1 to an unlimited number of
        siteType codes.  Default is to return all types.  List of valid site
        types: http://help.waterdata.usgs.gov/site_tp_cd::

            --siteType=ST       # Streams only
            --siteType=ST,LA-OU # Streams and Land Outcrops only""",
    'modifiedSince': r"""modifiedSince :
        [optional, default is None]

        Returns all values for sites and period of record requested only if any
        values have changed over the last modifiedSince period.  modifiedSince
        is useful if you periodically need to poll a site but are only
        interested in getting data if some of it has changed.  It is typically
        be used with period, or startDT/endDT but does not have to be. In the
        latter case, if any values were changed during the specified
        modifiedSince period, only the most recent values would be retrieved
        for those sites. This is a typical usage, since users typically are
        polling a site and only want data if there are new or changed
        measurements.  ISO-8601 duration format is always used.  There is no
        default.  (http://en.wikipedia.org/wiki/ISO_8601#Durations)::

            --modifiedSince=PT2H
                   # Retrieves all values for sites and period of record
                   # requested for any of the requested sites and parameters,
                   # but only for sites where any of the values changed during
                   # the last two hours.
            --modifiedSince=PT2H --period=P1D
                   # Retrieve all values for sites and period of record
                   # requested for the last 24 hours from now only for sites
                   # and parameters that had any values that changed or were
                   # added during the last two hours.
            --modifiedSince=PT2H --startDt=2010-11-01 --endDt=2010-11-02
                   # Retrieve all values for sites and period of record
                   # requested for sites and parameters that had values change
                   # between midnight site local time on Nov 1st, 2010 and
                   # 23:59 on Nov 2nd, 2010 site local time, only if values
                   # were changed or added within the last two hours.""",
    'agencyCd': r"""agencyCd :
        [optional, default is None]

        The list of sites returned are filtered to return only those with the
        provided agency code. The agency code describes the organization that
        maintains the site. Only one agency code is allowed and is optional.
        An authoritative list of agency codes can be found here.  Default is to
        return all sites regardless of agency code.  List:
        http://help.waterdata.usgs.gov/code/agency_cd_query?fmt=html::

            --stateCd=il --agencyCd=USCE # Only US Army Corps
                                         # of Engineers sites
                                         # in Illinois""",
    'siteStatus': r"""siteStatus :
        [optional, default is None]

        Selects sites based on whether or not they are active. If a site is
        active, it implies that it is being actively maintained. A site is
        considered active if: it has collected time-series (automated) data
        within the last 183 days (6 months), or it has collected discrete
        (manually collected) data within 397 days (13 months) If it does not
        meet these criteria, it is considered inactive. Some exceptions apply.
        If a site is flagged by a USGS water science center as discontinued, it
        will show as inactive. A USGS science center can also flag a new site
        as active even if it has not collected any data.  The default is all
        (show both active and inactive sites).  Chose between, 'all', 'active',
        or 'inactive'.  Default all - sites of any activity status are
        returned.::

            --siteStatus='active'""",
    'altMin': r"""altMin : float
        [optional, default is None]

        These arguments allows you to select instantaneous values sites where
        the associated sites' altitude are within a desired altitude, expressed
        in feet.  Altitude is based on the datum used at the site.  Providing
        a value to altMin (minimum altitude) means you want sites that have or
        exceed the altMin value.  You may specify decimal feet if precision is
        critical If both the altMin and altMax are specified, sites at or
        between the minimum and maximum altitude are returned.""",
    'altMax': r"""altMax : float
        [optional, default is None]

        Providing a value to altMax (maximum altitude) means you want sites
        that have or are less than the altMax value.::

            --altMin=1000 --altMax=5000
                  # Return sites where the altitude is 1000 feet or greater and
                  # 5000 feet or less.
            --altMin=12.5 --altMax=13
                  # Return sites where the altitude is 12.5 feet or greater and
                  # 13 feet or less.""",
    'drainAreaMin': r"""drainAreaMin : float
        [optional, default is None]

        SURFACE WATER SITE ATTRIBUTE

        These arguments allows you to select principally surface water sites
        where the associated sites' drainage areas (watersheds) are within
        a desired size, expressed in square miles or decimal fractions thereof.
        Providing a value to drainAreaMin (minimum drainage area) means you
        want sites that have or exceed the drainAreaMin value.  The values may
        be expressed in decimals. If both the drainAreaMin and drainAreaMax are
        specified, sites at or between the minimum and maximum drainage areas
        values specified are returned Caution: not all sites are associated
        with a drainage area.  Caution: drainage area generally only applies to
        surface water sites.  Use with other site types, such as groundwater
        sites, will likely retrieve no results.""",
    'drainAreaMax': r"""drainAreaMax:  float
        [optional, default is None]

        SURFACE WATER SITE ATTRIBUTE

        Providing a value to drainAreaMax (maximum drainage area) means you
        want sites that have or are less than the drainAreaMax value.::

            --drainAreaMin=1000 --drainAreaMax=5000
                                 # Return sites where the
                                 # drainage area is 1000
                                 # square miles or greater
                                 # and is 5000 square miles
                                 # or less.
            --drainAreaMin=10.5 --drainAreaMax=10.7
                                 # Return sites where the
                                 # drainage area is 10.5
                                 # square miles or greater
                                 # and is 10.7 square miles
                                 # or less.""",
    'aquiferCd': r"""aquiferCd
        [optional, default is None]

        Used to filter sites to those that exist in specified national
        aquifers. Note: not all sites have been associated with national
        aquifers.  Enter one or more national aquifer codes, separated by
        commas.  A national aquifer code is exactly 10 characters.  You can
        have up to 1000 aquiferCd codes.  Complete list:
        http://water.usgs.gov/ogw/NatlAqCode-reflist.html::

            --aquiferCd=S500EDRTRN,N100HGHPLN
                                  # returns groundwater sites
                                  # for the Edwards-Trinity
                                  # aquifer system and the
                                  # High Plains national
                                  # aquifers.""",
    'localAquiferCd': r"""localAquiferCd
        [optional, default is None]

        Used to filter sites to those that exist in specified local aquifers.
        Note: not all sites have been associated with local aquifers.  Enter
        one or more local aquifer codes, separated by commas.  A local aquifer
        code begins with a 2 character state abbreviation (such as TX for
        Texas) followed by a colon followed by the 7 character aquifer code.
        Can have 0 to 1000 comma delimited codes.  Complete list:
        http://help.waterdata.usgs.gov/code/aqfr_cd_query?fmt=html To translate
        state codes associated with the local aquifer you may need this
        reference: http://www.itl.nist.gov/fipspubs/fip5-2.htm ::

            --localAquiferCd=AL:111RGLT,AL:111RSDM
                    # returns sites for the Regolith and
                    # Saprolite local aquifers in Alabama""",
    'wellDepthMin': r"""wellDepthMin : float
        [optional, default is None]

        GROUNDWATER SITE ATTRIBUTE

        These arguments allows you to select groundwater sites serving data
        recorded automatically where the associated sites' well depth are
        within a desired depth, expressed in feet from the land surface datum.
        Express well depth as a positive number.  Providing a value to
        wellDepthMin (minimum well depth) means you want sites that have or
        exceed the wellDepthMin value.  The values may be expressed in decimals
        Caution: well depth applies to groundwater sites only.::

             --wellDepthMin=100 --wellDepthMax=500
                     # Return daily value sites where the well depth is 100
                     # feet or greater and 500 feet or less.""",
    'wellDepthMax': r"""wellDepthMax : float
        [optional, default is None]

        GROUNDWATER SITE ATTRIBUTE

        Providing a value to wellDepthMax (maximum well depth) means you want
        sites that have or are less than the wellDepthMax value.::

             --wellDepthMin=10.5 --wellDepthMax=10.7
                     # Return daily value sites where the well depth is 10.5
                     # feet or greater and 10.7 feet or less.

        If both the wellDepthMin and wellDepthMax are specified, sites at or
        between the minimum and maximum well depth values specified are
        returned wellDepthMax should be greater than or equal to
        wellDepthMin.""",
    'holeDepthMin': r"""holeDepthMin : float
        [optional, default is None]

        GROUNDWATER SITE ATTRIBUTE

        These arguments allows you to select groundwater sites serving data
        recorded automatically where the associated sites' hole depth are
        within a desired depth, expressed in feet from the land surface datum.
        Express hole depth as a positive number.  Providing a value to
        holeDepthMin (minimum hole depth) means you want sites that have or
        exceed the holeDepthMin value.  The values may be expressed in decimals
        Caution: hole depth applies to groundwater sites only.""",
    'holeDepthMax': r"""holeDepthMax : float
        [optional, default is None]

        GROUNDWATER SITE ATTRIBUTE

        Providing a value to holeDepthMax (maximum hole depth) means you want
        sites that have or are less than the holeDepthMax value.::

            --holeDepthMin=100 --holeDepthMax=500
                    # Return daily values sites where the hole depth is 100
                    # feet or greater and 500 feet or less.

            --holeDepthMin=10.5 --holeDepthMax=10.7
                    # Return daily value sites where the hole depth is 10.5
                    # feet or greater and 10.7 feet or less.

        If both the holeDepthMin and holeDepthMax are specified, sites at or
        between the minimum and maximum hole depth values specified are
        returned holeDepthMax should be greater than or equal to
        holeDepthMin.""",
    'period': r"""period
        [optional, default is None]

        Get a range of values from now by specifying the period argument period
        must be in ISO-8601 Duration format.
        (http://en.wikipedia.org/wiki/ISO_8601#Durations) Negative periods (ex:
        P-T2H) are not allowed.  Data are always returned up to the most recent
        value, which in the case of a predictive gage might be in the future.
        When specifying days from now, the first value will probably not be at
        midnight of the first day, but somewhat before exactly 24 hours from
        now.::

            --period=PT2H
                  # Retrieve last two hours from now up to most recent
                  # instantaneous value)
            --period=P7D
                  # Retrieve last seven days up from now to most recent
                  # instantaneous value)""",
    'startDT': r"""startDT
        [optional, default is None]

        Get a range of values from an explicit begin or end date/time.  Use the
        startDT and endDT arguments.  Site local time is output, even if
        multiple sites are requested and sites are in different time zones.
        Note that the measurement time zone at a site may not be the same as
        the time zone actually in effect at the site.

        Both startDt and endDt must be in ISO-8601 Date/Time format.
        (http://en.wikipedia.org/wiki/ISO_8601#Dates) You can express the date
        and time in a timezone other than site local time if you want as long
        as it follows the ISO standard. For example, you can express the time
        in Universal time: 2014-03-20T00:00Z.  If startDT is supplied and endDT
        is not, endDT ends with the most recent instantaneous value. startDT
        must be chronologically before endDT.

        If startDt shows the date and not the time of day (ex: 2010-09-01) the
        time of midnight site time is assumed (2010-09-01T00:00) If endDt shows
        the date and not the time of day (ex: 2010-09-02) the last minute
        before midnight site time is assumed (2010-09-02T23:59).  Remember,
        only data from October 1, 2007 are currently available in the 'iv'
        database.""",
    'endDT': r"""endDT
        [optional, default is None]

        If endDT is present, startDt must also be
        present.::

            --startDT=2010-11-22 --endDT=2010-11-22  # Full day, 00:00 to 23:59
            --startDT=2010-11-22T12:00 --endDT=2010-11-22T18:00
            --startDT=2010-11-22 --endDT=2010-11-22
            --startDT=2010-11-22T12:00  # From "startDT" to most recent
                                        # instantaneous value""",
    'statReportType': r"""statReportType : str
        [optional, default is 'daily']

        The type of statistics desired. Valid statistic report
        types include:

        +----------------+------------------------------------------+
        | statReportType | Description                              |
        +----------------+------------------------------------------+
        | daily          | daily statistics (default)               |
        |                | statistic across years                   |
        +----------------+------------------------------------------+
        | monthly        | monthly statistics (monthly time-series) |
        +----------------+------------------------------------------+
        | annual         | annual statistics, based on either       |
        |                | calendar year or water year, as defined  |
        |                | by statYearType. If statYearType is not  |
        |                | provided, calendar year statistics are   |
        |                | assumed. (annual time-series)            |
        +----------------+------------------------------------------+""",
    'statType': r"""statType : str
        [optional, default is None]

        Selects sites based on the statistics type(s) desired, such as minimum,
        maximum or mean

        For all statReportType types include::

            mean - arithmetic mean or average
            all - selects all available statistics

        For daily statistics you can also specify::

            min - minimum, or smallest value found for the
                  daily statistics
            max - maximum, or largest value found for the
                  daily statistics
            median - the numerical value separating the higher
                     half of a the data from the lower half,
                     same as specifying P50. If used median
                     will be represented by the column name
                     p50_va.
            P05, P10, P20, P25, P50, P75, P80, P90, P95
                with the number indicating percentile. Note:
                the service can calculate only these
                percentiles.""",
    'missingData': r"""missingData
        [optional, default is None]

        Used to indicate the rules to follow to generate statistics if there
        are gaps in the period of record during the requested statistics
        period. By default if there are any missing data for the report type,
        the statistic is left blank or null.

        This option does not apply to daily statistics, but optionally can be
        used with monthly and yearly statistics. If used with daily statistics,
        an error will occur.

        Missing data can happen for various reasons including there was
        a technical problem with the gage for part of the time period.

        Enabling this switch will attempt to provide a statistic if there is
        enough data to create one.

        Choice is 'off' or 'on'.""",
    'statisticsCd': r"""statisticsCd
        [optional, default is None]

        The statisticsCd represents how the instantaneous values are
        aggregated.  The statisticsCd js from the following table:

        +-------+------------------------------------+
        | Code  | Description                        |
        +=======+====================================+
        | 00001 | MAXIMUM VALUES                     |
        +-------+------------------------------------+
        | 00002 | MINIMUM VALUES                     |
        +-------+------------------------------------+
        | 00003 | MEAN VALUES                        |
        +-------+------------------------------------+
        | 00004 | VALUES TAKEN BETWEEN 0001 AND 1200 |
        +-------+------------------------------------+
        | 00005 | VALUES TAKEN BETWEEN 1201 AND 2400 |
        +-------+------------------------------------+
        | 00006 | SUMMATION VALUES                   |
        +-------+------------------------------------+
        | 00007 | MODAL VALUES                       |
        +-------+------------------------------------+
        | 00008 | MEDIAN VALUES                      |
        +-------+------------------------------------+
        | 00009 | STANDARD DEVIATION VALUES          |
        +-------+------------------------------------+
        | 00010 | VARIANCE VALUES                    |
        +-------+------------------------------------+
        | 00011 | RANDOM INSTANTANEOUS VALUES        |
        +-------+------------------------------------+
        | 00012 | EQUIVALENT MEAN VALUES             |
        +-------+------------------------------------+
        | 00013 | SKEWNESS VALUES                    |
        +-------+------------------------------------+
        | 00021 | TIDAL HIGH-HIGH VALUES             |
        +-------+------------------------------------+
        | 00022 | TIDAL LOW-HIGH VALUES              |
        +-------+------------------------------------+
        | 00023 | TIDAL HIGH-LOW VALUES              |
        +-------+------------------------------------+
        | 00024 | TIDAL LOW-LOW VALUES               |
        +-------+------------------------------------+
        | 01XXY | XX.Y PERCENTILE                    |
        +-------+------------------------------------+
        | 02LLL | LLL DAY LOW MEAN                   |
        +-------+------------------------------------+
        | 03HHH | HHH DAY HIGH MEAN                  |
        +-------+------------------------------------+
        | 3TTTT | INSTANTANEOUS OBSERVATION AT TTTT  |
        +-------+------------------------------------+""",
    'siteOutput': r"""siteOutput
        [optional, default is None]

        If you would like to see expanded site information, check this box.
        This argument is ignored for visually oriented output formats like
        Mapper, Google Earth and Google Maps. The default is basic. Use
        expanded to get expanded site information. Example:
        &siteOutput=expanded. Note: for performance reasons,
        &siteOutput=expanded cannot be used if seriesCatalogOutput=true or with
        any values for outputDataTypeCd.""",
    'seriesCatalogOutput': r"""seriesCatalogOutput
        [optional, default is None]

        This argument is ignored for visually oriented output formats like
        Mapper, Google Earth and Google Maps. If you would like to see all the
        period of record information for the sites selected, check this box.
        You will see detailed information, such as a continuous range of dates
        served by a site for one or more data types, for example, the begin and
        end dates that streamflow (parameter 00060) was recorded at a site.
        Note: if you select any data types for output (see below) the period of
        record data will also appear. In that case specifying this argument is
        unnecessary. The default is false. The only legal values for this
        argument are true and false. Example: &seriesCatalogOutput=true.
        &seriesCatalogOutput=true is equivalent to &outputDataTypeCd=all. Note:
        for performance reasons, &siteOutput=expanded cannot be used if
        seriesCatalogOutput=true.""",
    'outputDataTypeCd': r"""outputDataTypeCd
        [optional, default is None]

        This will add period of record information to certain output formats
        (GML, RDB and JSON) that summarize information about the data types
        requested.  The default is all data types. Some output formats are
        designed for visual use (Google Earth, Google Maps and Mapper).
        Consequently with these formats you will not see data type code
        information.

        Default information: If seriesCatalogOutput is true, all period of
        record information is shown by default. If seriesCatalogOutput is
        false, unless you override it using one of the values below, no period
        of record information is shown.

        Note: for performance reasons, &siteOutput=expanded cannot be used if
        with any values for outputDataTypeCd.

        Here are the various output data type codes available. These can be
        selected individually or can be added as comma separated values if
        desired.  Example: &outputDataTypeCd=iv,dv

        +-----+---------------------------------------------------------------+
        | all | default (see above for qualifications). This is equivalent to |
        |     | &seriesCatalogOutput=true.                                    |
        +-----+---------------------------------------------------------------+
        | iv  | Instantaneous values (time-series measurements typically      |
        |     | recorded by automated equipment at frequent intervals (e.g.,  |
        |     | hourly)                                                       |
        +-----+---------------------------------------------------------------+
        | uv  | Unit values (alias for iv)                                    |
        +-----+---------------------------------------------------------------+
        | rt  | Real-time data (alias for iv)                                 |
        +-----+---------------------------------------------------------------+
        | dv  | Daily values (once daily measurements or summarized           |
        |     | information for a particular day, such as daily maximum,      |
        |     | minimum and mean)                                             |
        +-----+---------------------------------------------------------------+
        | pk  | Peaks measurements of water levels and streamflow for surface |
        |     | water sites (such as during floods, may be either an          |
        |     | automated or a manual measurement)                            |
        +-----+---------------------------------------------------------------+
        | sv  | Site visits (irregular manual surface water measurements,     |
        |     | excluding peak measurements)                                  |
        +-----+---------------------------------------------------------------+
        | gw  | Groundwater levels measured at irregular, discrete intervals. |
        |     | For recorded, time series groundwater levels, use iv or id.   |
        +-----+---------------------------------------------------------------+
        | qw  | Water-quality data from discrete sampling events and analyzed |
        |     | in the field or in a laboratory. For recorded time series     |
        |     | water-quality data, use iv or id.                             |
        +-----+---------------------------------------------------------------+
        | id  | Historical instantaneous values (sites in the USGS            |
        |     | Instantaneous Data Archive External Link)                     |
        +-----+---------------------------------------------------------------+
        | aw  | Sites monitored by the USGS Active Groundwater Level Network  |
        |     | External Link                                                 |
        +-----+---------------------------------------------------------------+
        | ad  | Sites included in USGS Annual Water Data Reports External     |
        |     | Link}                                                         |
        +-----+---------------------------------------------------------------+""",
    'siteName': r"""siteName
        [optional, default is None]

        This filter allows you to find a site by its name, using either the
        exact site name or a partial site name. Note that a major filter is
        still required. String matches are case insensitive, so if you specify
        "Boulder" you will retrieve site names with "Boulder", "boulder",
        "BOULDER" as well as many other variants.  To embed a space, you can
        use single quotes. Examaple: --siteName='Boulder Creek'""",
    'siteNameMatchOperator': r"""siteNameMatchOperator
        [optional, default is None]

        If used, this must be used with siteName. It determines how the pattern
        matching for the site name behaves. Matches are case insensitive. The
        options are::

            start = The string must be at the start of the site name (default)
            any = The string must be contained somewhere in the site name
            exact = The site name must exactly match the string supplied, with
                    the exception that the match is not case sensitive

        Example: &siteNameMatchOperator=any""",
    'hasDataTypeCd': r"""hasDataTypeCd
        [optional, default is None]

        Default is all. Restricts results to those sites that collect certain
        kinds of data. Separate values with commas. Allowed values are:

        +-----+---------------------------------------------------------------+
        | all | default (see above for qualifications). This is equivalent to |
        |     | &seriesCatalogOutput=true.                                    |
        +-----+---------------------------------------------------------------+
        | iv  | Instantaneous values (time-series measurements typically      |
        |     | recorded by automated equipment at frequent intervals (e.g.,  |
        |     | hourly)                                                       |
        +-----+---------------------------------------------------------------+
        | uv  | Unit values (alias for iv)                                    |
        +-----+---------------------------------------------------------------+
        | rt  | Real-time data (alias for iv)                                 |
        +-----+---------------------------------------------------------------+
        | dv  | Daily values (once daily measurements or summarized           |
        |     | information for a particular day, such as daily maximum,      |
        |     | minimum and mean)                                             |
        +-----+---------------------------------------------------------------+
        | pk  | Peaks measurements of water levels and streamflow for surface |
        |     | water sites (such as during floods, may be either an          |
        |     | automated or a manual measurement)                            |
        +-----+---------------------------------------------------------------+
        | sv  | Site visits (irregular manual surface water measurements,     |
        |     | excluding peak measurements)                                  |
        +-----+---------------------------------------------------------------+
        | gw  | Groundwater levels measured at irregular, discrete intervals. |
        |     | For recorded, time series groundwater levels, use iv or id.   |
        +-----+---------------------------------------------------------------+
        | qw  | Water-quality data from discrete sampling events and analyzed |
        |     | in the field or in a laboratory. For recorded time series     |
        |     | water-quality data, use iv or id.                             |
        +-----+---------------------------------------------------------------+
        | id  | Historical instantaneous values (sites in the USGS            |
        |     | Instantaneous Data Archive External Link)                     |
        +-----+---------------------------------------------------------------+
        | aw  | Sites monitored by the USGS Active Groundwater Level Network  |
        |     | External Link                                                 |
        +-----+---------------------------------------------------------------+
        | ad  | Sites included in USGS Annual Water Data Reports External     |
        |     | Link                                                          |
        +-----+---------------------------------------------------------------+""",
    'statYearType': r"""statYearType
        [optional, default is None]

        Indicates which kind of year statistics should be created against. This
        only applies when requesting annual statistics, i.e.
        statReportType=annual. Valid year types codes include:

        +----------+----------------------------------------------------------+
        | calendar | calendar year, i.e. January 1 through December 31        |
        +----------+----------------------------------------------------------+
        | water    | water year, i.e. a year begins October 1 of the previous |
        |          | year and ends September 30 of the current year. This is  |
        |          | the same as a federal fiscal year.                       |
        +----------+----------------------------------------------------------+"""}


@mando.command(formatter_class=HelpFormatter, doctype='numpy')
def nwis(sites=None,
         stateCd=None,
         huc=None,
         bBox=None,
         countyCd=None,
         parameterCd=None,
         period=None,
         startDT=None,
         endDT=None,
         siteType=None,
         modifiedSince=None,
         agencyCd=None,
         siteStatus=None,
         altMin=None,
         altMax=None,
         drainAreaMin=None,
         drainAreaMax=None,
         aquiferCd=None,
         localAquiferCd=None,
         wellDepthMin=None,
         wellDepthMax=None,
         holeDepthMin=None,
         holeDepthMax=None,
         database='dv',
         statReportType=None,
         statType=None,
         missingData=None,
         statYearType=None):
    r"""Deprecated: Use the ``nwis_*`` functions instead.

    This function has been split up into individual functions for each source
    database.  This allows for keywords and output to be tailored to each
    specific web service.

    +-------------------+-------------------------------+
    | New Function      | Database Name and Description |
    +===================+===============================+
    | nwis_iv           | Instantaneous Value sub-daily |
    |                   | interval data starting in     |
    |                   | 2007                          |
    +-------------------+-------------------------------+
    | nwis_dv           | Daily Values database         |
    +-------------------+-------------------------------+
    | nwis_stat         | daily/monthly/annual          |
    |                   | statistics                    |
    +-------------------+-------------------------------+
    | nwis_site         | Site metadata                 |
    +-------------------+-------------------------------+
    | nwis_measurements | Field measurements            |
    +-------------------+-------------------------------+
    | nwis_peak         | Peak flow and stage           |
    +-------------------+-------------------------------+
    | nwis_gwlevels     | Ground water levels           |
    +-------------------+-------------------------------+

    This function/sub-command will continue to work, however you should change
    all scripts to use the split out functions.
    """
    from tsgettoolbox.services.usgs import nwis as placeholder
    if database not in ['iv',
                        'dv',
                        'stat',
                        'measurements',
                        'peak',
                        'site',
                        'gwlevels']:
        raise ValueError(r"""
*
*   The 'database' option must be either 'iv' for instantaneous values,
*   or 'dv' for daily values, or 'stat' for daily, monthly, or annual
*   statistics, or 'measurements' for field measurements, 'peak' for
*   peak stage and flow estimates, 'site' for site metadata, or 'gwlevels' for
*   ground water levels.
*   You gave {0}.
*
""".format(database))
    url = r'http://waterservices.usgs.gov/nwis/{0}/'.format(database)
    if database in ['measurements', 'peak', 'gwlevels']:
        words = sites.split(',')
        if len(words) != 1:
            raise ValueError(r"""
*
*   For the 'measurements', 'peak', and 'gwlevels' databases you can only
*   collect data from one site, you listed {0}.
*
""".format(len(words)))
        if (stateCd is not None or
            huc is not None or
            bBox is not None or
                countyCd is not None):
            raise ValueError(r"""
*
*   The 'measurements', 'peak', or 'gwlevels' databases can currently only
*   accept one site using the 'site' keyword.
*
""")

        if database in ['measurements', 'peak']:
            url = r'http://nwis.waterdata.usgs.gov/XX/nwis/{0}'.format(database)
    r = resource(url,
                 sites=sites,
                 stateCd=stateCd,
                 huc=huc,
                 bBox=bBox,
                 countyCd=countyCd,
                 parameterCd=parameterCd,
                 siteType=siteType,
                 modifiedSince=modifiedSince,
                 agencyCd=agencyCd,
                 siteStatus=siteStatus,
                 altMin=altMin,
                 altMax=altMax,
                 drainAreaMin=drainAreaMin,
                 drainAreaMax=drainAreaMax,
                 aquiferCd=aquiferCd,
                 localAquiferCd=localAquiferCd,
                 wellDepthMin=wellDepthMin,
                 wellDepthMax=wellDepthMax,
                 holeDepthMin=holeDepthMin,
                 holeDepthMax=holeDepthMax,
                 period=period,
                 startDT=startDT,
                 endDT=endDT,
                 statReportType=statReportType,
                 statType=statType,
                 missingData=missingData,
                 statYearType=statYearType)

    return tsutils.printiso(odo(r, pd.DataFrame))


@mando.command('nwis_iv', formatter_class=HelpFormatter, doctype='numpy')
@tsutils.doc(nwis_docstrings)
def nwis_iv_cli(sites=None,
                stateCd=None,
                huc=None,
                bBox=None,
                countyCd=None,
                parameterCd=None,
                period=None,
                startDT=None,
                endDT=None,
                siteType=None,
                modifiedSince=None,
                agencyCd=None,
                siteStatus=None,
                altMin=None,
                altMax=None,
                drainAreaMin=None,
                drainAreaMax=None,
                aquiferCd=None,
                localAquiferCd=None,
                wellDepthMin=None,
                wellDepthMax=None,
                holeDepthMin=None,
                holeDepthMax=None,
                includeCodes=False):
    r"""Download from Instantaneous Values of the USGS NWIS.

    Detailed documentation is available at http://waterdata.usgs.gov/nwis.

    Site local time is output, even if multiple sites are requested and sites
    are in different time zones.  Note that the measurement time zone at a site
    may not be the same as the time zone actually in effect at the site.

    Every query requires a major filter. Pick the major filter ('--sites',
    '--stateCd', '--huc', '--bBox', '--countyCd') that best retrieves the data
    for the sites that you are interested in.  You can have only one major
    filter per query. If you specify more than one major filter, you will get
    an error.

    **Major Filter**

    Select ONE of::

        '--sites',
        '--stateCd',
        '--huc',
        '--bBox', or
        '--countyCd'

    **Minor Filters**

    Additional filters can be applied after specifying a major filter. This
    further reduces the set of expected results. Users are encouraged to use
    minor filters because it allows more efficient use of this service.

    Use as many as desired to limit number of retrieved time series::

        '--parameterCd',
        '--siteType',
        '--modifiedSince',
        '--agencyCd',
        '--siteStatus',
        '--altMin',
        '--altMax',
        '--drainAreaMin',
        '--drainAreaMax',
        '--aquiferCd',
        '--localAquiferCd',
        '--wellDepthMin',
        '--wellDepthMax',
        '--holeDepthMin',
        '--holeDepthMax'

    **Results**

    The column name in the resulting table is made up of
    "agencyCd_siteno_parameterCd", for example "USGS_02248380_00010".  The
    agency and parameter codes are described in the `agencyCd` and
    `parameterCd` options below.

    If `includeCodes` option is used, there will also be columns representing
    the data quality codes named "agencyCd_siteno_parameterCd_cd".

    +---------+--------------------------------------------------------+
    | Code    |  Description                                           |
    +=========+========================================================+
    | e       | Value has been edited or estimated by USGS personnel   |
    |         | and is write protected                                 |
    +---------+--------------------------------------------------------+
    | &       | Value was computed from affected unit values           |
    +---------+--------------------------------------------------------+
    | E       | Value was computed from estimated unit values.         |
    +---------+--------------------------------------------------------+
    | A       | Approved for publication -- Processing and review      |
    |         | completed.                                             |
    +---------+--------------------------------------------------------+
    | P       | Provisional data subject to revision.                  |
    +---------+--------------------------------------------------------+
    | <       | The value is known to be less than reported value and  |
    |         | is write protected.                                    |
    +---------+--------------------------------------------------------+
    | >       | The value is known to be greater than reported value   |
    |         | and is write protected.                                |
    +---------+--------------------------------------------------------+
    | 1       | Value is write protected without any remark code to be |
    |         | printed                                                |
    +---------+--------------------------------------------------------+
    | 2       | Remark is write protected without any remark code to   |
    |         | be printed                                             |
    +---------+--------------------------------------------------------+
    |         | No remark (blank)                                      |
    +---------+--------------------------------------------------------+
    | Ssn     | Parameter monitored seasonally                         |
    +---------+--------------------------------------------------------+
    | Ice     | Ice affected                                           |
    +---------+--------------------------------------------------------+
    | Pr      | Partial-record site                                    |
    +---------+--------------------------------------------------------+
    | Rat     | Rating being developed or revised                      |
    +---------+--------------------------------------------------------+
    | Eqp     | Equipment malfunction                                  |
    +---------+--------------------------------------------------------+
    | Fld     | Flood damage                                           |
    +---------+--------------------------------------------------------+
    | Dis     | Data-collection discontinued                           |
    +---------+--------------------------------------------------------+
    | Dry     | Dry                                                    |
    +---------+--------------------------------------------------------+
    | --      | Parameter not determined                               |
    +---------+--------------------------------------------------------+
    | Mnt     | Maintenance in progress                                |
    +---------+--------------------------------------------------------+
    | ZFl     | Zero flow                                              |
    +---------+--------------------------------------------------------+
    | ``***`` | Temporarily unavailable                                |
    +---------+--------------------------------------------------------+

    Parameters
    ----------
    {sites}
    {huc}
    {bBox}
    {countyCd}
    {parameterCd}
    {agencyCd}
    {stateCd}
    {altMin}
    {altMax}
    {aquiferCd}
    {endDT}
    {localAquiferCd}
    {modifiedSince}
    {parameterCd}
    {period}
    {siteStatus}
    {siteType}
    {startDT}
    {drainAreaMin}
    {drainAreaMax}
    {holeDepthMin}
    {holeDepthMax}
    {wellDepthMin}
    {wellDepthMax}
    {includeCodes}"""
    tsutils._printiso(nwis_iv(sites=sites,
                              stateCd=stateCd,
                              huc=huc,
                              bBox=bBox,
                              countyCd=countyCd,
                              parameterCd=parameterCd,
                              siteType=siteType,
                              modifiedSince=modifiedSince,
                              agencyCd=agencyCd,
                              siteStatus=siteStatus,
                              altMin=altMin,
                              altMax=altMax,
                              drainAreaMin=drainAreaMin,
                              drainAreaMax=drainAreaMax,
                              aquiferCd=aquiferCd,
                              localAquiferCd=localAquiferCd,
                              wellDepthMin=wellDepthMin,
                              wellDepthMax=wellDepthMax,
                              holeDepthMin=holeDepthMin,
                              holeDepthMax=holeDepthMax,
                              period=period,
                              startDT=startDT,
                              endDT=endDT,
                              includeCodes=includeCodes))


def nwis_iv(sites=None,
            stateCd=None,
            huc=None,
            bBox=None,
            countyCd=None,
            parameterCd=None,
            period=None,
            startDT=None,
            endDT=None,
            siteType=None,
            modifiedSince=None,
            agencyCd=None,
            siteStatus=None,
            altMin=None,
            altMax=None,
            drainAreaMin=None,
            drainAreaMax=None,
            aquiferCd=None,
            localAquiferCd=None,
            wellDepthMin=None,
            wellDepthMax=None,
            holeDepthMin=None,
            holeDepthMax=None,
            includeCodes=False):
    r"""Download from Instantaneous Values of the USGS NWIS."""
    from tsgettoolbox.services.usgs import nwis as placeholder
    url = r'http://waterservices.usgs.gov/nwis/iv/'
    r = resource(url,
                 sites=sites,
                 stateCd=stateCd,
                 huc=huc,
                 bBox=bBox,
                 countyCd=countyCd,
                 parameterCd=parameterCd,
                 siteType=siteType,
                 modifiedSince=modifiedSince,
                 agencyCd=agencyCd,
                 siteStatus=siteStatus,
                 altMin=altMin,
                 altMax=altMax,
                 drainAreaMin=drainAreaMin,
                 drainAreaMax=drainAreaMax,
                 aquiferCd=aquiferCd,
                 localAquiferCd=localAquiferCd,
                 wellDepthMin=wellDepthMin,
                 wellDepthMax=wellDepthMax,
                 holeDepthMin=holeDepthMin,
                 holeDepthMax=holeDepthMax,
                 period=period,
                 startDT=startDT,
                 endDT=endDT,
                 includeCodes=includeCodes)

    return odo(r, pd.DataFrame)


@mando.command('nwis_dv', formatter_class=HelpFormatter, doctype='numpy')
@tsutils.doc(nwis_docstrings)
def nwis_dv_cli(sites=None,
                stateCd=None,
                huc=None,
                bBox=None,
                countyCd=None,
                parameterCd=None,
                statisticsCd=None,
                period=None,
                startDT=None,
                endDT=None,
                siteType=None,
                modifiedSince=None,
                agencyCd=None,
                siteStatus=None,
                altMin=None,
                altMax=None,
                drainAreaMin=None,
                drainAreaMax=None,
                aquiferCd=None,
                localAquiferCd=None,
                wellDepthMin=None,
                wellDepthMax=None,
                holeDepthMin=None,
                holeDepthMax=None,
                includeCodes=False):
    r"""Download from the Daily Values database of the USGS NWIS.

    Detailed documentation is available at http://waterdata.usgs.gov/nwis.

    Site local time is output, even if multiple sites are requested and sites
    are in different time zones.  Note that the measurement time zone at a site
    may not be the same as the time zone actually in effect at the site.

    Every query requires a major filter. Pick the major filter ('--sites',
    '--stateCd', '--huc', '--bBox', '--countyCd') that best retrieves the data
    for the sites that you are interested in.  You can have only one major
    filter per query. If you specify more than one major filter, you will get
    an error.

    **Major Filter**

    Select one of::

        '--sites',
        '--stateCd',
        '--huc',
        '--bBox', or
        '--countyCd'

    **Minor Filters**

    Additional filters can be applied after specifying a major filter. This
    further reduces the set of expected results. Users are encouraged to use
    minor filters because it allows more efficient use of this service.

    Use as many as desired to limit number of retrieved time series::

        '--parameterCd',
        '--siteType',
        '--modifiedSince',
        '--agencyCd',
        '--siteStatus',
        '--altMin',
        '--altMax',
        '--drainAreaMin',
        '--drainAreaMax',
        '--aquiferCd',
        '--localAquiferCd',
        '--wellDepthMin',
        '--wellDepthMax',
        '--holeDepthMin',
        '--holeDepthMax'

    **Results**

    The column name in the resulting table is made up of
    "USGS_SITE_CODE-parameterCd", for example
    "02248380-00010".  The parameter code is described in the
    "parameterCd" option below.

    If `includeCodes` option is used, there will also be columns representing
    the data quality codes named "agencyCd_siteno_parameterCd_cd".

    +---------+--------------------------------------------------------+
    | Code    |  Description                                           |
    +=========+========================================================+
    | e       | Value has been edited or estimated by USGS personnel   |
    |         | and is write protected                                 |
    +---------+--------------------------------------------------------+
    | &       | Value was computed from affected unit values           |
    +---------+--------------------------------------------------------+
    | E       | Value was computed from estimated unit values.         |
    +---------+--------------------------------------------------------+
    | A       | Approved for publication -- Processing and review      |
    |         | completed.                                             |
    +---------+--------------------------------------------------------+
    | P       | Provisional data subject to revision.                  |
    +---------+--------------------------------------------------------+
    | <       | The value is known to be less than reported value and  |
    |         | is write protected.                                    |
    +---------+--------------------------------------------------------+
    | >       | The value is known to be greater than reported value   |
    |         | and is write protected.                                |
    +---------+--------------------------------------------------------+
    | 1       | Value is write protected without any remark code to be |
    |         | printed                                                |
    +---------+--------------------------------------------------------+
    | 2       | Remark is write protected without any remark code to   |
    |         | be printed                                             |
    +---------+--------------------------------------------------------+
    |         | No remark (blank)                                      |
    +---------+--------------------------------------------------------+
    | Ssn     | Parameter monitored seasonally                         |
    +---------+--------------------------------------------------------+
    | Ice     | Ice affected                                           |
    +---------+--------------------------------------------------------+
    | Pr      | Partial-record site                                    |
    +---------+--------------------------------------------------------+
    | Rat     | Rating being developed or revised                      |
    +---------+--------------------------------------------------------+
    | Eqp     | Equipment malfunction                                  |
    +---------+--------------------------------------------------------+
    | Fld     | Flood damage                                           |
    +---------+--------------------------------------------------------+
    | Dis     | Data-collection discontinued                           |
    +---------+--------------------------------------------------------+
    | Dry     | Dry                                                    |
    +---------+--------------------------------------------------------+
    | --      | Parameter not determined                               |
    +---------+--------------------------------------------------------+
    | Mnt     | Maintenance in progress                                |
    +---------+--------------------------------------------------------+
    | ZFl     | Zero flow                                              |
    +---------+--------------------------------------------------------+
    | ``***`` | Temporarily unavailable                                |
    +---------+--------------------------------------------------------+

    Parameters
    ----------
    {sites}
    {huc}
    {bBox}
    {countyCd}
    {agencyCd}
    {stateCd}
    {altMin}
    {altMax}
    {aquiferCd}
    {endDT}
    {localAquiferCd}
    {modifiedSince}
    {parameterCd}
    {statisticsCd}
    {period}
    {siteStatus}
    {siteType}
    {startDT}
    {drainAreaMin}
    {drainAreaMax}
    {holeDepthMin}
    {holeDepthMax}
    {wellDepthMin}
    {wellDepthMax}
    {includeCodes}"""
    tsutils._printiso(nwis_dv(sites=sites,
                              stateCd=stateCd,
                              huc=huc,
                              bBox=bBox,
                              countyCd=countyCd,
                              parameterCd=parameterCd,
                              statisticsCd=statisticsCd,
                              siteType=siteType,
                              modifiedSince=modifiedSince,
                              agencyCd=agencyCd,
                              siteStatus=siteStatus,
                              altMin=altMin,
                              altMax=altMax,
                              drainAreaMin=drainAreaMin,
                              drainAreaMax=drainAreaMax,
                              aquiferCd=aquiferCd,
                              localAquiferCd=localAquiferCd,
                              wellDepthMin=wellDepthMin,
                              wellDepthMax=wellDepthMax,
                              holeDepthMin=holeDepthMin,
                              holeDepthMax=holeDepthMax,
                              period=period,
                              startDT=startDT,
                              endDT=endDT,
                              includeCodes=includeCodes))


def nwis_dv(sites=None,
            stateCd=None,
            huc=None,
            bBox=None,
            countyCd=None,
            parameterCd=None,
            statisticsCd=None,
            period=None,
            startDT=None,
            endDT=None,
            siteType=None,
            modifiedSince=None,
            agencyCd=None,
            siteStatus=None,
            altMin=None,
            altMax=None,
            drainAreaMin=None,
            drainAreaMax=None,
            aquiferCd=None,
            localAquiferCd=None,
            wellDepthMin=None,
            wellDepthMax=None,
            holeDepthMin=None,
            holeDepthMax=None,
            includeCodes=False):
    r"""Download from the Daily Values database of the USGS NWIS."""
    from tsgettoolbox.services.usgs import nwis as placeholder
    url = r'http://waterservices.usgs.gov/nwis/dv/'
    r = resource(url,
                 sites=sites,
                 stateCd=stateCd,
                 huc=huc,
                 bBox=bBox,
                 countyCd=countyCd,
                 parameterCd=parameterCd,
                 statisticsCd=statisticsCd,
                 siteType=siteType,
                 modifiedSince=modifiedSince,
                 agencyCd=agencyCd,
                 siteStatus=siteStatus,
                 altMin=altMin,
                 altMax=altMax,
                 drainAreaMin=drainAreaMin,
                 drainAreaMax=drainAreaMax,
                 aquiferCd=aquiferCd,
                 localAquiferCd=localAquiferCd,
                 wellDepthMin=wellDepthMin,
                 wellDepthMax=wellDepthMax,
                 holeDepthMin=holeDepthMin,
                 holeDepthMax=holeDepthMax,
                 period=period,
                 startDT=startDT,
                 endDT=endDT,
                 includeCodes=includeCodes)

    return odo(r, pd.DataFrame)


@mando.command('nwis_site', formatter_class=HelpFormatter, doctype='numpy')
@tsutils.doc(nwis_docstrings)
def nwis_site_cli(sites=None,
                  stateCd=None,
                  huc=None,
                  bBox=None,
                  countyCd=None,
                  parameterCd=None,
                  period=None,
                  startDT=None,
                  endDT=None,
                  siteType=None,
                  modifiedSince=None,
                  agencyCd=None,
                  siteStatus=None,
                  altMin=None,
                  altMax=None,
                  drainAreaMin=None,
                  drainAreaMax=None,
                  aquiferCd=None,
                  localAquiferCd=None,
                  wellDepthMin=None,
                  wellDepthMax=None,
                  holeDepthMin=None,
                  holeDepthMax=None,
                  siteOutput=None,
                  seriesCatalogOutput=None,
                  outputDataTypeCd=None,
                  siteName=None,
                  siteNameMatchOperator=None,
                  hasDataTypeCd=None):
    r"""Download from the site database of the USGS NWIS.

    Detailed documentation is available at http://waterdata.usgs.gov/nwis.

    This does not return a time-series, but a table of sites.

    Every query requires a major filter. Pick the major filter ('--sites',
    '--stateCd', '--huc', '--bBox', '--countyCd') that best retrieves the data
    for the sites that you are interested in.  You can have only one major
    filter per query. If you specify more than one major filter, you will get
    an error.

    **Major Filter**

    Select one of::

        '--sites',
        '--stateCd',
        '--huc',
        '--bBox', or
        '--countyCd'

    **Minor Filters**

    Additional filters can be applied after specifying a major filter. This
    further reduces the set of expected results. Users are encouraged to use
    minor filters because it allows more efficient use of this service.

    Use as many as desired to limit number of retrieved time series::

        '--parameterCd',
        '--siteType',
        '--modifiedSince',
        '--agencyCd',
        '--siteStatus',
        '--altMin',
        '--altMax',
        '--drainAreaMin',
        '--drainAreaMax',
        '--aquiferCd',
        '--localAquiferCd',
        '--wellDepthMin',
        '--wellDepthMax',
        '--holeDepthMin',
        '--holeDepthMax'

    **Results**

    +---------------------+----------------------------------+
    | Column              | Description                      |
    +=====================+==================================+
    | agency_cd           | Agency                           |
    +---------------------+----------------------------------+
    | site_no             | Site identification number       |
    +---------------------+----------------------------------+
    | station_nm          | Site name                        |
    +---------------------+----------------------------------+
    | site_tp_cd          | Site type                        |
    +---------------------+----------------------------------+
    | dec_lat_va          | Decimal latitude                 |
    +---------------------+----------------------------------+
    | dec_long_va         | Decimal longitude                |
    +---------------------+----------------------------------+
    | coord_acy_cd        | Latitude-longitude accuracy      |
    +---------------------+----------------------------------+
    | dec_coord_datum_cd  | Decimal Latitude-longitude datum |
    +---------------------+----------------------------------+
    | alt_va              | Altitude of Gage/land surface    |
    +---------------------+----------------------------------+
    | alt_acy_va          | Altitude accuracy                |
    +---------------------+----------------------------------+
    | alt_datum_cd        | Altitude datum                   |
    +---------------------+----------------------------------+
    | huc_cd              | Hydrologic unit code             |
    +---------------------+----------------------------------+

    .. _site_tp_cd: https://help.waterdata.usgs.gov/code/site_tp_query?fmt=html
    .. _coord_acy_cd: https://help.waterdata.usgs.gov/code/coord_acy_cd_query?fmt=html
    .. _dec_coord_datum_cd: https://help.waterdata.usgs.gov/code/coord_datum_cd_query?fmt=html
    .. _alt_datum_cd: https://help.waterdata.usgs.gov/code/alt_datum_cd_query?fmt=html
    .. _huc_cd: https://help.waterdata.usgs.gov/code/hucs_query?fmt=html

    Parameters
    ----------
    {sites}
    {huc}
    {bBox}
    {countyCd}
    {agencyCd}
    {stateCd}
    {altMin}
    {altMax}
    {aquiferCd}
    {endDT}
    {localAquiferCd}
    {modifiedSince}
    {parameterCd}
    {statisticsCd}
    {period}
    {siteStatus}
    {siteType}
    {startDT}
    {drainAreaMin}
    {drainAreaMax}
    {holeDepthMin}
    {holeDepthMax}
    {wellDepthMin}
    {wellDepthMax}
    {siteOutput}
    {seriesCatalogOutput}
    {outputDataTypeCd}
    {siteName}
    {siteNameMatchOperator}
    {hasDataTypeCd}"""
    tsutils._printiso(nwis_site(sites=sites,
                                stateCd=stateCd,
                                huc=huc,
                                bBox=bBox,
                                countyCd=countyCd,
                                parameterCd=parameterCd,
                                siteType=siteType,
                                modifiedSince=modifiedSince,
                                agencyCd=agencyCd,
                                siteStatus=siteStatus,
                                altMin=altMin,
                                altMax=altMax,
                                drainAreaMin=drainAreaMin,
                                drainAreaMax=drainAreaMax,
                                aquiferCd=aquiferCd,
                                localAquiferCd=localAquiferCd,
                                wellDepthMin=wellDepthMin,
                                wellDepthMax=wellDepthMax,
                                holeDepthMin=holeDepthMin,
                                holeDepthMax=holeDepthMax,
                                period=period,
                                startDT=startDT,
                                endDT=endDT,
                                siteOutput=siteOutput,
                                seriesCatalogOutput=seriesCatalogOutput,
                                outputDataTypeCd=outputDataTypeCd,
                                siteName=siteName,
                                siteNameMatchOperator=siteNameMatchOperator,
                                hasDataTypeCd=hasDataTypeCd))


def nwis_site(sites=None,
              stateCd=None,
              huc=None,
              bBox=None,
              countyCd=None,
              parameterCd=None,
              period=None,
              startDT=None,
              endDT=None,
              siteType=None,
              modifiedSince=None,
              agencyCd=None,
              siteStatus=None,
              altMin=None,
              altMax=None,
              drainAreaMin=None,
              drainAreaMax=None,
              aquiferCd=None,
              localAquiferCd=None,
              wellDepthMin=None,
              wellDepthMax=None,
              holeDepthMin=None,
              holeDepthMax=None,
              siteOutput=None,
              seriesCatalogOutput=None,
              outputDataTypeCd=None,
              siteName=None,
              siteNameMatchOperator=None,
              hasDataTypeCd=None):
    r"""Download from the site database of the USGS NWIS."""
    from tsgettoolbox.services.usgs import nwis as placeholder
    url = r'http://waterservices.usgs.gov/nwis/site/'
    r = resource(url,
                 sites=sites,
                 stateCd=stateCd,
                 huc=huc,
                 bBox=bBox,
                 countyCd=countyCd,
                 parameterCd=parameterCd,
                 siteType=siteType,
                 modifiedSince=modifiedSince,
                 agencyCd=agencyCd,
                 siteStatus=siteStatus,
                 altMin=altMin,
                 altMax=altMax,
                 drainAreaMin=drainAreaMin,
                 drainAreaMax=drainAreaMax,
                 aquiferCd=aquiferCd,
                 localAquiferCd=localAquiferCd,
                 wellDepthMin=wellDepthMin,
                 wellDepthMax=wellDepthMax,
                 holeDepthMin=holeDepthMin,
                 holeDepthMax=holeDepthMax,
                 period=period,
                 startDT=startDT,
                 endDT=endDT,
                 siteOutput=siteOutput,
                 seriesCatalogOutput=seriesCatalogOutput,
                 outputDataTypeCd=outputDataTypeCd,
                 siteName=siteName,
                 siteNameMatchOperator=siteNameMatchOperator,
                 hasDataTypeCd=hasDataTypeCd)

    return odo(r, pd.DataFrame)


@mando.command('nwis_gwlevels', formatter_class=HelpFormatter, doctype='numpy')
@tsutils.doc(nwis_docstrings)
def nwis_gwlevels_cli(sites=None,
                      stateCd=None,
                      huc=None,
                      bBox=None,
                      countyCd=None,
                      parameterCd=None,
                      period=None,
                      startDT=None,
                      endDT=None,
                      siteType=None,
                      modifiedSince=None,
                      agencyCd=None,
                      siteStatus=None,
                      altMin=None,
                      altMax=None,
                      aquiferCd=None,
                      localAquiferCd=None,
                      wellDepthMin=None,
                      wellDepthMax=None,
                      holeDepthMin=None,
                      holeDepthMax=None):
    r"""Download from the Groundwater Levels database of the USGS NWIS.

    Detailed documentation is available at http://waterdata.usgs.gov/nwis.

    Site local time is output, even if multiple sites are requested and sites
    are in different time zones.  Note that the measurement time zone at a site
    may not be the same as the time zone actually in effect at the site.

    Every query requires a major filter. Pick the major filter ('--sites',
    '--stateCd', '--huc', '--bBox', '--countyCd') that best retrieves the data
    for the sites that you are interested in.  You can have only one major
    filter per query. If you specify more than one major filter, you will get
    an error.

    **Major Filter**

    Select one of::

        '--sites',
        '--stateCd',
        '--huc',
        '--bBox', or
        '--countyCd'

    **Minor Filters**

    Additional filters can be applied after specifying a major filter. This
    further reduces the set of expected results. Users are encouraged to use
    minor filters because it allows more efficient use of this service.

    Use as many as desired to limit number of retrieved time series::

        '--parameterCd',
        '--siteType',
        '--modifiedSince',
        '--agencyCd',
        '--siteStatus',
        '--altMin',
        '--altMax',
        '--drainAreaMin',
        '--drainAreaMax',
        '--aquiferCd',
        '--localAquiferCd',
        '--wellDepthMin',
        '--wellDepthMax',
        '--holeDepthMin',
        '--holeDepthMax'

    **Results**

    +---------------+-------------------------------+
    | Column        | Description                   |
    +===============+===============================+
    | agency_cd     | Agency code                   |
    +---------------+-------------------------------+
    | site_no       | USGS site number              |
    +---------------+-------------------------------+
    | site_tp_cd    | Site type code                |
    +---------------+-------------------------------+
    | lev_dt        | Date level measured           |
    +---------------+-------------------------------+
    | lev_tm        | Time level measured           |
    +---------------+-------------------------------+
    | lev_tz_cd     | Time datum                    |
    +---------------+-------------------------------+
    | lev_va        | Water-level value in feet     |
    |               | below land surface            |
    +---------------+-------------------------------+
    | sl_lev_va     | Water-level value in feet     |
    |               | above specific vertical datum |
    +---------------+-------------------------------+
    | sl_datum_cd   | Referenced vertical datum     |
    +---------------+-------------------------------+
    | lev_status_cd | Status                        |
    +---------------+-------------------------------+
    | lev_agency_cd | Measuring agency              |
    +---------------+-------------------------------+

    Parameters
    ----------
    {sites}
    {huc}
    {bBox}
    {countyCd}
    {agencyCd}
    {stateCd}
    {altMin}
    {altMax}
    {aquiferCd}
    {endDT}
    {localAquiferCd}
    {modifiedSince}
    {parameterCd}
    {period}
    {siteStatus}
    {siteType}
    {startDT}
    {holeDepthMin}
    {holeDepthMax}
    {wellDepthMin}
    {wellDepthMax}"""
    tsutils._printiso(nwis_gwlevels(sites=sites,
                                    stateCd=stateCd,
                                    huc=huc,
                                    bBox=bBox,
                                    countyCd=countyCd,
                                    parameterCd=parameterCd,
                                    siteType=siteType,
                                    modifiedSince=modifiedSince,
                                    agencyCd=agencyCd,
                                    siteStatus=siteStatus,
                                    altMin=altMin,
                                    altMax=altMax,
                                    aquiferCd=aquiferCd,
                                    localAquiferCd=localAquiferCd,
                                    wellDepthMin=wellDepthMin,
                                    wellDepthMax=wellDepthMax,
                                    holeDepthMin=holeDepthMin,
                                    holeDepthMax=holeDepthMax,
                                    period=period,
                                    startDT=startDT,
                                    endDT=endDT))


def nwis_gwlevels(sites=None,
                  stateCd=None,
                  huc=None,
                  bBox=None,
                  countyCd=None,
                  parameterCd=None,
                  period=None,
                  startDT=None,
                  endDT=None,
                  siteType=None,
                  modifiedSince=None,
                  agencyCd=None,
                  siteStatus=None,
                  altMin=None,
                  altMax=None,
                  aquiferCd=None,
                  localAquiferCd=None,
                  wellDepthMin=None,
                  wellDepthMax=None,
                  holeDepthMin=None,
                  holeDepthMax=None):
    from tsgettoolbox.services.usgs import nwis as placeholder
    url = r'http://waterservices.usgs.gov/nwis/gwlevels/'
    r = resource(url,
                 sites=sites,
                 stateCd=stateCd,
                 huc=huc,
                 bBox=bBox,
                 countyCd=countyCd,
                 parameterCd=parameterCd,
                 siteType=siteType,
                 modifiedSince=modifiedSince,
                 agencyCd=agencyCd,
                 siteStatus=siteStatus,
                 altMin=altMin,
                 altMax=altMax,
                 aquiferCd=aquiferCd,
                 localAquiferCd=localAquiferCd,
                 wellDepthMin=wellDepthMin,
                 wellDepthMax=wellDepthMax,
                 holeDepthMin=holeDepthMin,
                 holeDepthMax=holeDepthMax,
                 period=period,
                 startDT=startDT,
                 endDT=endDT)

    return odo(r, pd.DataFrame)


@mando.command('nwis_measurements', formatter_class=HelpFormatter, doctype='numpy')
@tsutils.doc(nwis_docstrings)
def nwis_measurements_cli(sites=None,
                          stateCd=None,
                          huc=None,
                          bBox=None,
                          countyCd=None,
                          parameterCd=None,
                          period=None,
                          startDT=None,
                          endDT=None,
                          siteType=None,
                          modifiedSince=None,
                          agencyCd=None,
                          siteStatus=None,
                          altMin=None,
                          altMax=None,
                          aquiferCd=None,
                          localAquiferCd=None,
                          wellDepthMin=None,
                          wellDepthMax=None,
                          holeDepthMin=None,
                          holeDepthMax=None):
    r"""Download from the Measurements database of the USGS NWIS.

    Detailed documentation is available at http://waterdata.usgs.gov/nwis.

    Site local time is output, even if multiple sites are requested and sites
    are in different time zones.  Note that the measurement time zone at a site
    may not be the same as the time zone actually in effect at the site.

    Every query requires a major filter. Pick the major filter ('--sites',
    '--stateCd', '--huc', '--bBox', '--countyCd') that best retrieves the data
    for the sites that you are interested in.  You can have only one major
    filter per query. If you specify more than one major filter, you will get
    an error.

    **Major Filter**

    Select one of::

        '--sites',
        '--stateCd',
        '--huc',
        '--bBox', or
        '--countyCd'

    **Minor Filters**

    Additional filters can be applied after specifying a major filter. This
    further reduces the set of expected results. Users are encouraged to use
    minor filters because it allows more efficient use of this service.

    Use as many as desired to limit number of retrieved time series::

        '--parameterCd',
        '--siteType',
        '--modifiedSince',
        '--agencyCd',
        '--siteStatus',
        '--altMin',
        '--altMax',
        '--drainAreaMin',
        '--drainAreaMax',
        '--aquiferCd',
        '--localAquiferCd',
        '--wellDepthMin',
        '--wellDepthMax',
        '--holeDepthMin',
        '--holeDepthMax'

    **Results**

    +---------------------------+-------------------------------------------+
    | Column                    | Description                               |
    +===========================+===========================================+
    | agency_cd                 | Agency code                               |
    +---------------------------+-------------------------------------------+
    | site_no                   | Site number                               |
    +---------------------------+-------------------------------------------+
    | measurement_nu            | Measurement number                        |
    +---------------------------+-------------------------------------------+
    | measurement_dt            | date of measurement (format = MMDDYYYY    |
    |                           | or Month/Day/Year. The user has options   |
    |                           | for the data output format)               |
    +---------------------------+-------------------------------------------+
    | tz_cd                     | Time zone offset. An ANSI SQL/92 time     |
    |                           | zone offset string. Some examples are     |
    |                           | '-07:00' (Eastern), '+02:00' (Eastern     |
    |                           | Europe), and '+03:30' (India).            |
    +---------------------------+-------------------------------------------+
    | q_meas_used_fg            | Flag indicates if the discharge           |
    |                           | measurement is marked used.               |
    +---------------------------+-------------------------------------------+
    | party_nm                  | an indication of who made the             |
    |                           | measurement and is usually populated with |
    |                           | a pair of initials separated with a slash |
    +---------------------------+-------------------------------------------+
    | site_visit_coll_agency_cd | What agency made the measurement at the   |
    |                           | site visit                                |
    +---------------------------+-------------------------------------------+
    | gage_height_va            | gage height as shown on the inside staff  |
    |                           | gage at the site or read off the recorder |
    |                           | inside the gage house in feet             |
    +---------------------------+-------------------------------------------+
    | discharge_va              | the computed discharge in cubic feet per  |
    |                           | second (cfs)                              |
    +---------------------------+-------------------------------------------+
    | measured_rating_diff      | measurement rating codes that denote the  |
    |                           | relative quality of the measurement       |
    +---------------------------+-------------------------------------------+
    | gage_va_change            | The amount the gage height changed while  |
    |                           | the measurement was being made in feet    |
    +---------------------------+-------------------------------------------+
    | gage_va_time              | The amount of time elapsed while the      |
    |                           | measurement was being made in decimal     |
    |                           | hours                                     |
    +---------------------------+-------------------------------------------+
    | control_type_cd           | condition of the rating control at the    |
    |                           | time of the measurement                   |
    +---------------------------+-------------------------------------------+
    | discharge_cd              | The adjustment code for the measured      |
    |                           | discharge                                 |
    +---------------------------+-------------------------------------------+
    | chan_nu                   | The channel number                        |
    +---------------------------+-------------------------------------------+
    | chan_name                 | The channel name                          |
    +---------------------------+-------------------------------------------+
    | meas_type                 | The channel measurement type              |
    +---------------------------+-------------------------------------------+
    | streamflow_method         | The channel discharge measurement method  |
    +---------------------------+-------------------------------------------+
    | velocity_method           | The channel velocity measurement method   |
    +---------------------------+-------------------------------------------+
    | chan_discharge            | The channel discharge in cubic feet per   |
    |                           | second                                    |
    +---------------------------+-------------------------------------------+
    | chan_width                | The channel width in feet                 |
    +---------------------------+-------------------------------------------+
    | chan_area                 | The channel area in square feet           |
    +---------------------------+-------------------------------------------+
    | chan_velocity             | The mean velocity in feet per second      |
    +---------------------------+-------------------------------------------+
    | chan_stability            | The stability of the channel material     |
    +---------------------------+-------------------------------------------+
    | chan_material             | The channel material                      |
    +---------------------------+-------------------------------------------+
    | chan_evenness             | The channel evenness from bank to bank    |
    +---------------------------+-------------------------------------------+
    | long_vel_desc             | The longitudinal velocity description     |
    +---------------------------+-------------------------------------------+
    | horz_vel_desc             | The horizontal velocity description       |
    +---------------------------+-------------------------------------------+
    | vert_vel_desc             | The vertical velocity description         |
    +---------------------------+-------------------------------------------+
    | chan_loc_cd               | The channel location code                 |
    +---------------------------+-------------------------------------------+
    | chan_loc_dist             | The channel location distance             |
    +---------------------------+-------------------------------------------+

    https://help.waterdata.usgs.gov/output-formats#streamflow_measurement_data

    Parameters
    ----------
    {sites}
    {huc}
    {bBox}
    {countyCd}
    {agencyCd}
    {stateCd}
    {altMin}
    {altMax}
    {aquiferCd}
    {endDT}
    {localAquiferCd}
    {modifiedSince}
    {parameterCd}
    {period}
    {siteStatus}
    {siteType}
    {startDT}
    {holeDepthMin}
    {holeDepthMax}
    {wellDepthMin}
    {wellDepthMax}"""
    tsutils._printiso(nwis_measurements_cli(sites=sites,
                                            stateCd=stateCd,
                                            huc=huc,
                                            bBox=bBox,
                                            countyCd=countyCd,
                                            parameterCd=parameterCd,
                                            siteType=siteType,
                                            modifiedSince=modifiedSince,
                                            agencyCd=agencyCd,
                                            siteStatus=siteStatus,
                                            altMin=altMin,
                                            altMax=altMax,
                                            aquiferCd=aquiferCd,
                                            localAquiferCd=localAquiferCd,
                                            wellDepthMin=wellDepthMin,
                                            wellDepthMax=wellDepthMax,
                                            holeDepthMin=holeDepthMin,
                                            holeDepthMax=holeDepthMax,
                                            period=period,
                                            startDT=startDT,
                                            endDT=endDT))


def nwis_measurements(sites=None,
                      stateCd=None,
                      huc=None,
                      bBox=None,
                      countyCd=None,
                      parameterCd=None,
                      period=None,
                      startDT=None,
                      endDT=None,
                      siteType=None,
                      modifiedSince=None,
                      agencyCd=None,
                      siteStatus=None,
                      altMin=None,
                      altMax=None,
                      aquiferCd=None,
                      localAquiferCd=None,
                      wellDepthMin=None,
                      wellDepthMax=None,
                      holeDepthMin=None,
                      holeDepthMax=None):
    r"""Download from the Measurements database of the USGS NWIS."""
    from tsgettoolbox.services.usgs import nwis as placeholder
    url = r'http://nwis.waterdata.usgs.gov/XX/nwis/measurements'
    r = resource(url,
                 sites=sites,
                 stateCd=stateCd,
                 huc=huc,
                 bBox=bBox,
                 countyCd=countyCd,
                 parameterCd=parameterCd,
                 siteType=siteType,
                 modifiedSince=modifiedSince,
                 agencyCd=agencyCd,
                 siteStatus=siteStatus,
                 altMin=altMin,
                 altMax=altMax,
                 aquiferCd=aquiferCd,
                 localAquiferCd=localAquiferCd,
                 wellDepthMin=wellDepthMin,
                 wellDepthMax=wellDepthMax,
                 holeDepthMin=holeDepthMin,
                 holeDepthMax=holeDepthMax,
                 period=period,
                 startDT=startDT,
                 endDT=endDT)

    return odo(r, pd.DataFrame)


@mando.command('nwis_peak', formatter_class=HelpFormatter, doctype='numpy')
@tsutils.doc(nwis_docstrings)
def nwis_peak_cli(sites=None,
                  stateCd=None,
                  huc=None,
                  bBox=None,
                  countyCd=None,
                  parameterCd=None,
                  period=None,
                  startDT=None,
                  endDT=None,
                  siteType=None,
                  modifiedSince=None,
                  agencyCd=None,
                  siteStatus=None,
                  altMin=None,
                  altMax=None,
                  aquiferCd=None,
                  localAquiferCd=None,
                  wellDepthMin=None,
                  wellDepthMax=None,
                  holeDepthMin=None,
                  holeDepthMax=None):
    r"""Download from the Peak database of the USGS NWIS.

    Detailed documentation is available at http://waterdata.usgs.gov/nwis.

    Site local time is output, even if multiple sites are requested and sites
    are in different time zones.  Note that the measurement time zone at a site
    may not be the same as the time zone actually in effect at the site.

    Every query requires a major filter. Pick the major filter ('--sites',
    '--stateCd', '--huc', '--bBox', '--countyCd') that best retrieves the data
    for the sites that you are interested in.  You can have only one major
    filter per query. If you specify more than one major filter, you will get
    an error.

    **Major Filter**

    Select one of::

        '--sites',
        '--stateCd',
        '--huc',
        '--bBox', or
        '--countyCd'

    **Minor Filters**

    Additional filters can be applied after specifying a major filter. This
    further reduces the set of expected results. Users are encouraged to use
    minor filters because it allows more efficient use of this service.

    Use as many as desired to limit number of retrieved time series::

        '--parameterCd',
        '--siteType',
        '--modifiedSince',
        '--agencyCd',
        '--siteStatus',
        '--altMin',
        '--altMax',
        '--drainAreaMin',
        '--drainAreaMax',
        '--aquiferCd',
        '--localAquiferCd',
        '--wellDepthMin',
        '--wellDepthMax',
        '--holeDepthMin',
        '--holeDepthMax'

    **Results**

    +---------------+--------------------------------------------+
    | Column        | Description                                |
    +===============+============================================+
    | agency_cd     | Agency Code                                |
    +---------------+--------------------------------------------+
    | site_no       | USGS station number                        |
    +---------------+--------------------------------------------+
    | peak_dt       | Date of peak streamflow                    |
    |               |   (format YYYY-MM-DD)                      |
    +---------------+--------------------------------------------+
    | peak_tm       | Time of peak streamflow                    |
    |               |   (24 hour format, 00:00 - 23:59)          |
    +---------------+--------------------------------------------+
    | peak_va       | Annual peak streamflow value in cfs        |
    +---------------+--------------------------------------------+
    | peak_cd       | Peak Discharge-Qualification codes         |
    |               |   (see explanation below)                  |
    +---------------+--------------------------------------------+
    | gage_ht       | Gage height for the associated peak        |
    |               |   streamflow in feet                       |
    +---------------+--------------------------------------------+
    | gage_ht_cd    | Gage height qualification codes            |
    +---------------+--------------------------------------------+
    | year_last_pk  | Peak streamflow reported is the highest    |
    |               |   since this year                          |
    +---------------+--------------------------------------------+
    | ag_dt         | Date of maximum gage-height for water year |
    |               |   (if not concurrent with peak)            |
    +---------------+--------------------------------------------+
    | ag_tm         | Time of maximum gage-height for water year |
    |               |   (if not concurrent with peak)            |
    +---------------+--------------------------------------------+
    | ag_gage_ht    | maximum Gage height for water year in feet |
    |               |   (if not concurrent with peak)            |
    +---------------+--------------------------------------------+
    | ag_gage_ht_cd | maximum Gage height code                   |
    +---------------+--------------------------------------------+

    Peak Streamflow-Qualification Codes(peak_cd):
    +---------+---------------------------------------------------+
    | peak_cd | Description                                          |
    +=========+======================================================+
    | 1       | Discharge is a Maximum Daily Average                 |
    +---------+------------------------------------------------------+
    | 2       | Discharge is an Estimate                             |
    +---------+------------------------------------------------------+
    | 3       | Discharge affected by Dam Failure                    |
    +---------+------------------------------------------------------+
    | 4       | Discharge less than indicated value,                 |
    |         |   which is Minimum Recordable Discharge at this site |
    +---------+------------------------------------------------------+
    | 5       | Discharge affected to unknown degree by              |
    |         |   Regulation or Diversion                            |
    +---------+------------------------------------------------------+
    | 6       | Discharge affected by Regulation or Diversion        |
    +---------+------------------------------------------------------+
    | 7       | Discharge is an Historic Peak                        |
    +---------+------------------------------------------------------+
    | 8       | Discharge actually greater than indicated value      |
    +---------+------------------------------------------------------+
    | 9       | Discharge due to Snowmelt, Hurricane,                |
    |         |   Ice-Jam or Debris Dam breakup                      |
    +---------+------------------------------------------------------+
    | A       | Year of occurrence is unknown or not exact           |
    +---------+------------------------------------------------------+
    | B       | Month or Day of occurrence is unknown or not exact   |
    +---------+------------------------------------------------------+
    | C       | All or part of the record affected by Urbanization,  |
    |         |    Mining, Agricultural changes, Channelization,     |
    |         |    or other                                          |
    +---------+------------------------------------------------------+
    | D       | Base Discharge changed during this year              |
    +---------+------------------------------------------------------+
    | E       | Only Annual Maximum Peak available for this year     |
    +---------+------------------------------------------------------+

    Gage height qualification codes(gage_ht_cd,ag_gage_ht_cd):
    +---------------+------------------------------------------------+
    | gage_ht_cd    | Description                                    |
    | ag_gage_ht_cd |                                                |
    +===============+================================================+
    | 1             | Gage height affected by backwater              |
    +---------------+------------------------------------------------+
    | 2             | Gage height not the maximum for the year       |
    +---------------+------------------------------------------------+
    | 3             | Gage height at different site and(or) datum    |
    +---------------+------------------------------------------------+
    | 4             | Gage height below minimum recordable elevation |
    +---------------+------------------------------------------------+
    | 5             | Gage height is an estimate                     |
    +---------------+------------------------------------------------+
    | 6             | Gage datum changed during this year            |
    +---------------+------------------------------------------------+

    Parameters
    ----------
    {sites}
    {huc}
    {bBox}
    {countyCd}
    {agencyCd}
    {stateCd}
    {altMin}
    {altMax}
    {aquiferCd}
    {endDT}
    {localAquiferCd}
    {modifiedSince}
    {parameterCd}
    {period}
    {siteStatus}
    {siteType}
    {startDT}
    {holeDepthMin}
    {holeDepthMax}
    {wellDepthMin}
    {wellDepthMax}"""
    tsutils._printiso(nwis_peak(sites=sites,
                                stateCd=stateCd,
                                huc=huc,
                                bBox=bBox,
                                countyCd=countyCd,
                                parameterCd=parameterCd,
                                siteType=siteType,
                                modifiedSince=modifiedSince,
                                agencyCd=agencyCd,
                                siteStatus=siteStatus,
                                altMin=altMin,
                                altMax=altMax,
                                aquiferCd=aquiferCd,
                                localAquiferCd=localAquiferCd,
                                wellDepthMin=wellDepthMin,
                                wellDepthMax=wellDepthMax,
                                holeDepthMin=holeDepthMin,
                                holeDepthMax=holeDepthMax,
                                period=period,
                                startDT=startDT,
                                endDT=endDT))


def nwis_peak(sites=None,
              stateCd=None,
              huc=None,
              bBox=None,
              countyCd=None,
              parameterCd=None,
              period=None,
              startDT=None,
              endDT=None,
              siteType=None,
              modifiedSince=None,
              agencyCd=None,
              siteStatus=None,
              altMin=None,
              altMax=None,
              aquiferCd=None,
              localAquiferCd=None,
              wellDepthMin=None,
              wellDepthMax=None,
              holeDepthMin=None,
              holeDepthMax=None):
    r"""Download from the Peak database of the USGS NWIS."""
    from tsgettoolbox.services.usgs import nwis as placeholder
    url = r'http://nwis.waterdata.usgs.gov/XX/nwis/peak'
    r = resource(url,
                 sites=sites,
                 stateCd=stateCd,
                 huc=huc,
                 bBox=bBox,
                 countyCd=countyCd,
                 parameterCd=parameterCd,
                 siteType=siteType,
                 modifiedSince=modifiedSince,
                 agencyCd=agencyCd,
                 siteStatus=siteStatus,
                 altMin=altMin,
                 altMax=altMax,
                 aquiferCd=aquiferCd,
                 localAquiferCd=localAquiferCd,
                 wellDepthMin=wellDepthMin,
                 wellDepthMax=wellDepthMax,
                 holeDepthMin=holeDepthMin,
                 holeDepthMax=holeDepthMax,
                 period=period,
                 startDT=startDT,
                 endDT=endDT)

    return odo(r, pd.DataFrame)


@mando.command('nwis_stat', formatter_class=HelpFormatter, doctype='numpy')
@tsutils.doc(nwis_docstrings)
def nwis_stat_cli(sites=None,
                  parameterCd=None,
                  period=None,
                  startDT=None,
                  endDT=None,
                  siteType=None,
                  modifiedSince=None,
                  agencyCd=None,
                  siteStatus=None,
                  altMin=None,
                  altMax=None,
                  drainAreaMin=None,
                  drainAreaMax=None,
                  aquiferCd=None,
                  localAquiferCd=None,
                  wellDepthMin=None,
                  wellDepthMax=None,
                  holeDepthMin=None,
                  holeDepthMax=None,
                  statReportType=None,
                  statType=None,
                  missingData=None,
                  statYearType=None):
    r"""Download from the Statistic database of the USGS NWIS.

    Detailed documentation is available at http://waterdata.usgs.gov/nwis.

    Every query requires a major filter. Pick the major filter ('--sites',
    '--stateCd', '--huc', '--bBox', '--countyCd') that best retrieves the data
    for the sites that you are interested in.  You can have only one major
    filter per query. If you specify more than one major filter, you will get
    an error.

    **Major Filter**

    Select one of::

        '--sites',
        '--stateCd',
        '--huc',
        '--bBox', or
        '--countyCd'

    **Minor Filters**

    Additional filters can be applied after specifying a major filter. This
    further reduces the set of expected results. Users are encouraged to use
    minor filters because it allows more efficient use of this service.

    Use as many as desired to limit number of retrieved time series::

        '--parameterCd',
        '--siteType',
        '--modifiedSince',
        '--agencyCd',
        '--siteStatus',
        '--altMin',
        '--altMax',
        '--drainAreaMin',
        '--drainAreaMax',
        '--aquiferCd',
        '--localAquiferCd',
        '--wellDepthMin',
        '--wellDepthMax',
        '--holeDepthMin',
        '--holeDepthMax'

    **Returns**

    This returns a pandas DataFrame if using the Python API, or a text table to
    standard out if using the command line interface.  Both will have columns
    named according to the following table.

    +--------------+----------------+-----------------------------------------+
    | Column Name  | statReportType | Description                             |
    +==============+================+=========================================+
    | agency_cd    | all            | Agency code                             |
    +--------------+----------------+-----------------------------------------+
    | site_no      | all            | Site identification number              |
    +--------------+----------------+-----------------------------------------+
    | parameter_cd | all            | Parameter code                          |
    +--------------+----------------+-----------------------------------------+
    | station_nm   | all            | Site name                               |
    +--------------+----------------+-----------------------------------------+
    | loc_web_ds   | all            | Additional measurement description      |
    +--------------+----------------+-----------------------------------------+
    | year_nu      | monthly        | The year for which the statistics       |
    |              | annual         | apply.                                  |
    +--------------+----------------+-----------------------------------------+
    | month_nu     | monthly        | The month for which the statistics      |
    |              | daily          | apply.                                  |
    +--------------+----------------+-----------------------------------------+
    | day_nu       | daily          | The day for which the statistics apply. |
    +--------------+----------------+-----------------------------------------+
    | begin_yr     | daily          | First water year of data of daily mean  |
    |              |                | values for this day.                    |
    +--------------+----------------+-----------------------------------------+
    | end_yr       | daily          | Last water year of data of daily mean   |
    |              |                | values for this day.                    |
    +--------------+----------------+-----------------------------------------+
    | count_nu     | all            | Number of values used in the            |
    |              |                | calculation.                            |
    +--------------+----------------+-----------------------------------------+
    | max_va_yr    | daily          | Water year in which the maximum value   |
    |              |                | occurred.                               |
    +--------------+----------------+-----------------------------------------+
    | max_va       | daily          | Maximum of daily mean values for        |
    |              |                | this day.                               |
    +--------------+----------------+-----------------------------------------+
    | min_va_yr    | daily          | Water year in which the minimum value   |
    |              |                | occurred.                               |
    +--------------+----------------+-----------------------------------------+
    | min_va       | daily          | Minimum of daily mean values for        |
    |              |                | this day.                               |
    +--------------+----------------+-----------------------------------------+
    | mean_va      | all            | Mean of daily mean values for this day. |
    +--------------+----------------+-----------------------------------------+
    | p05_va       | daily          | 05 percentile of daily mean values for  |
    |              |                | this day.                               |
    +--------------+----------------+-----------------------------------------+
    | p10_va       | daily          | 10 percentile of daily mean values for  |
    |              |                | this day.                               |
    +--------------+----------------+-----------------------------------------+
    | p20_va       | daily          | 20 percentile of daily mean values for  |
    |              |                | this day.                               |
    +--------------+----------------+-----------------------------------------+
    | p25_va       | daily          | 25 percentile of daily mean values for  |
    |              |                | this day.                               |
    +--------------+----------------+-----------------------------------------+
    | p50_va       | daily          | 50 percentile (median) of daily         |
    |              |                | mean values for this day.               |
    +--------------+----------------+-----------------------------------------+
    | p75_va       | daily          | 75 percentile of daily mean values for  |
    |              |                | this day.                               |
    +--------------+----------------+-----------------------------------------+
    | p80_va       | daily          | 80 percentile of daily mean values for  |
    |              |                | this day.                               |
    +--------------+----------------+-----------------------------------------+
    | p90_va       | daily          | 90 percentile of daily mean values for  |
    |              |                | this day.                               |
    +--------------+----------------+-----------------------------------------+
    | p95_va       | daily          | 95 percentile of daily mean values for  |
    |              |                | this day.                               |
    +--------------+----------------+-----------------------------------------+

    Parameters
    ----------
    {sites}
    {agencyCd}
    {altMin}
    {altMax}
    {aquiferCd}
    {endDT}
    {localAquiferCd}
    {modifiedSince}
    {parameterCd}
    {period}
    {siteStatus}
    {siteType}
    {startDT}
    {drainAreaMin}
    {drainAreaMax}
    {holeDepthMin}
    {holeDepthMax}
    {wellDepthMin}
    {wellDepthMax}
    {statReportType}
    {statType}
    {missingData}
    {statYearType}"""
    tsutils._printiso(nwis_stat(sites=sites,
                                parameterCd=parameterCd,
                                siteType=siteType,
                                modifiedSince=modifiedSince,
                                agencyCd=agencyCd,
                                siteStatus=siteStatus,
                                altMin=altMin,
                                altMax=altMax,
                                drainAreaMin=drainAreaMin,
                                drainAreaMax=drainAreaMax,
                                aquiferCd=aquiferCd,
                                localAquiferCd=localAquiferCd,
                                wellDepthMin=wellDepthMin,
                                wellDepthMax=wellDepthMax,
                                holeDepthMin=holeDepthMin,
                                holeDepthMax=holeDepthMax,
                                period=period,
                                startDT=startDT,
                                endDT=endDT,
                                statReportType=statReportType,
                                statType=statType,
                                missingData=missingData,
                                statYearType=statYearType))


def nwis_stat(sites=None,
              parameterCd=None,
              period=None,
              startDT=None,
              endDT=None,
              siteType=None,
              modifiedSince=None,
              agencyCd=None,
              siteStatus=None,
              altMin=None,
              altMax=None,
              drainAreaMin=None,
              drainAreaMax=None,
              aquiferCd=None,
              localAquiferCd=None,
              wellDepthMin=None,
              wellDepthMax=None,
              holeDepthMin=None,
              holeDepthMax=None,
              statReportType=None,
              statType=None,
              missingData=None,
              statYearType=None):
    r"""Download from the Statistic database of the USGS NWIS."""
    from tsgettoolbox.services.usgs import nwis as placeholder
    url = r'http://waterservices.usgs.gov/nwis/stat/'
    r = resource(url,
                 sites=sites,
                 parameterCd=parameterCd,
                 siteType=siteType,
                 modifiedSince=modifiedSince,
                 agencyCd=agencyCd,
                 siteStatus=siteStatus,
                 altMin=altMin,
                 altMax=altMax,
                 drainAreaMin=drainAreaMin,
                 drainAreaMax=drainAreaMax,
                 aquiferCd=aquiferCd,
                 localAquiferCd=localAquiferCd,
                 wellDepthMin=wellDepthMin,
                 wellDepthMax=wellDepthMax,
                 holeDepthMin=holeDepthMin,
                 holeDepthMax=holeDepthMax,
                 period=period,
                 startDT=startDT,
                 endDT=endDT,
                 statReportType=statReportType,
                 statType=statType,
                 missingData=missingData,
                 statYearType=statYearType,
                 )

    return odo(r, pd.DataFrame)


@mando.command('epa_wqp', formatter_class=HelpFormatter, doctype='numpy')
@tsutils.doc(nwis_docstrings)
def epa_wqp_cli(bBox=None,
                lat=None,
                lon=None,
                within=None,
                countrycode=None,
                statecode=None,
                countycode=None,
                siteType=None,
                organization=None,
                siteid=None,
                huc=None,
                sampleMedia=None,
                characteristicType=None,
                characteristicName=None,
                pCode=None,
                activityId=None,
                startDateLo=None,
                startDateHi=None):
    r"""Download from the EPA Water Quality Portal.

    All of the keywords could be considered as subtractive filters of the
    stations.  Some of the keywords can result in very large number of stations
    being downloaded.

    Parameters
    ----------
    {bBox}
    lat : float
        [optional, default is None]

        Latitude for radial search, expressed in decimal degrees, WGS84

        The `lat`, `lon`, and `within` arguments are used together to form
        a circle on the Earth's surface for locating data-collection stations.
        Many stations outside the continental US do not have latitude and
        longitude referenced to WGS84 and therefore cannot be found using these
        parameters.

    lon : float
        [optional, default is None]

        Longitude for radial search, expressed in decimal degrees, WGS84

        The `lat`, `lon`, and `within` arguments are used together to form
        a circle on the Earth's surface for locating data-collection stations.
        Many stations outside the continental US do not have latitude and
        longitude referenced to WGS84 and therefore cannot be found using these
        parameters.

    within : float
        [optional, default is None]

        Distance for radial search, expressed in decimal miles

        The `lat`, `lon`, and `within` arguments are used together to form
        a circle on the Earth's surface for locating data-collection stations.
        Many stations outside the continental US do not have latitude and
        longitude referenced to WGS84 and therefore cannot be found using these
        parameters.

    countrycode : str
        [optional, default is None]

        Two-character Federal Information Processing Standard (FIPS) country
        code. (see domain service for available codes)

        FIPS country codes were established by the National Institute of
        Standards, publication 5-2.

    statecode : str
        [optional, default is None]

        Two-character Federal Information Processing Standard (FIPS) country
        code, followed bu a ":", followed by a two-digit FIPS state code. (see
        domain service for available codes)

        FIPS state codes were established by the National Institute of
        Standards, publication 5-2.

    countycode : str
        [optional, default is None]

        Two-character Federal Information Processing Standard (FIPS) country
        code, followed by a ":", followed by a two-digit FIPS state code,
        followed by a ":", followed by a three-digit FIPS county code. (see
        domain service for available codes)

        FIPS county codes were established by the National Institute of
        Standards, publication 5-2.

    siteType : str
        [optional, default is None]

        One or more case-sensitive site types, separated by semicolons. (see
        domain service for available site types)

        Restrict retrieval to stations with specified site type (location in
        the hydrologic cycle).  The MonitoringLocationTypeName for individual
        records may provide more detailed information about the type of
        individual stations.

    organization : str
        [optional, default is None]

        For USGS organization IDs, append an upper-case postal-service state
        abbreviation to "USGS-" to identify the USGS office managing the data
        collection station records. However, a few US states are serviced by
        one USGS office.::

            USGS-MA = Massachusetts and Rhode Island
            USGS-MD = Maryland, Delaware, and the District of Columbia
            USGS-PR = Caribbean Islands
            USGS-HI = Pacific Islands

        (see domain service for available organization IDs)

        USGS offices sometimes provide data for stations outside the political
        boundaries associated with the office's organization code. Use the
        statecode or countycode arguments to search for stations located within
        those political boundaries.

    siteid : str
        [optional, default is None]

        Concatenate an agency code, a hyphen ("-"), and a site-identification
        number.

        Each data collection station is assigned a unique
        site-identification number. Other agencies often use different site
        identification numbers for the same stations.

    {huc}

    sampleMedia : str
        [optional, default is None]

        One or more case-sensitive sample media, separated by semicolons. (see
        domain service for available sample media)

        Sample media are broad general classes, and may be subdivided in the
        retrieved data. Examine the data elements ActivityMediaName,
        ActivityMediaSubdivisionName, and ResultSampleFractionText for more
        detailed information.

    characteristicType : str
        [optional, default is None]

        One or more case-sensitive characteristic types (groupings) separated
        by semicolons. (see domain service for available characteristic types)

        These groups will be expanded as part of the ongoing collaboration
        between USGS and USEPA.

    characteristicName : str
        [optional, default is None]

        One or more case-sensitive characteristic names, separated by
        semicolons. (see domain service for available characteristic names)

        Characteristic names identify different types of environmental
        measurements. The names are derived from the USEPA Substance Registry
        System (SRS). USGS uses parameter codes for the same purpose and has
        associated most parameters to SRS names.

    pCode : str
        [optional, default is None]

        One or more five-digit USGS parameter codes, separated by semicolons.
        This is equivalent to "parameterCd" used in other USGS web services.

    activityId : str
        [optional, default is None]

        One or more case-sensitive activity IDs, separated by semicolons.
        Designator that uniquely identifies an activity within an organization.

    startDateLo : str
        [optional, default is None]

        Date of earliest desired data-collection activity.  A very wide range
        of date strings can be used but the closer to ISO 8601 the better.

    startDateHi : str
        [optional, default is None]

        Date of last desired data-collection activity.  A very wide range of
        date strings can be used but the closer to ISO 8601 the better."""
    tsutils._printiso(epa_wqp(bBox=bBox,
                              lat=lat,
                              lon=lon,
                              within=within,
                              countrycode=countrycode,
                              statecode=statecode,
                              countycode=countycode,
                              siteType=siteType,
                              organization=organization,
                              siteid=siteid,
                              huc=huc,
                              sampleMedia=sampleMedia,
                              characteristicType=characteristicType,
                              characteristicName=characteristicName,
                              pCode=pCode,
                              activityId=activityId,
                              startDateLo=startDateLo,
                              startDateHi=startDateHi))


def epa_wqp(bBox=None,
            lat=None,
            lon=None,
            within=None,
            countrycode=None,
            statecode=None,
            countycode=None,
            siteType=None,
            organization=None,
            siteid=None,
            huc=None,
            sampleMedia=None,
            characteristicType=None,
            characteristicName=None,
            pCode=None,
            activityId=None,
            startDateLo=None,
            startDateHi=None):
    from tsgettoolbox.services.epa import wqp as placeholder

    url = r'https://www.waterqualitydata.us/Result/search'
    r = resource(url,
                 bBox=bBox,
                 lat=lat,
                 lon=lon,
                 within=within,
                 countrycode=countrycode,
                 statecode=statecode,
                 countycode=countycode,
                 siteType=siteType,
                 organization=organization,
                 siteid=siteid,
                 huc=huc,
                 sampleMedia=sampleMedia,
                 characteristicType=characteristicType,
                 characteristicName=characteristicName,
                 pCode=pCode,
                 activityId=activityId,
                 startDateLo=startDateLo,
                 startDateHi=startDateHi,
                 )

    return odo(r, pd.DataFrame)




import mando

try:
    from mando.rst_text_formatter import RSTHelpFormatter as HelpFormatter
except ImportError:
    from argparse import RawTextHelpFormatter as HelpFormatter

from tstoolbox import tsutils


@mando.command('lcra_hydromet', formatter_class=HelpFormatter, doctype='numpy')
@tsutils.doc(tsutils.docstrings)
def lcra_hydromet_cli(site_code,
                      parameter_code,
                      start_date=None,
                      end_date=None,
                      dam_site_location='head'):
    r"""Hydrometeorologic data from the Lower Colorado River Authority.

    This module provides access to hydrologic and climate data in the Colorado
    River Basin (Texas) provided by the Lower Colorado River Authority Hydromet
    web site and web service.

    http://www.lcra.org

    http://hydromet.lcra.org

    Parameters
    ----------
    site_code
        The LCRA site code (four chars long) of the site you want to query data
        for.
    parameter_code
        LCRA parameter
        code.

        +----------------+---------------------------------------+
        | parameter_code | Description                           |
        +----------------+---------------------------------------+
        | stage          | the level of water above a benchmark  |
        |                | in feet                               |
        +----------------+---------------------------------------+
        | flow           | streamflow in cubic feet per second   |
        +----------------+---------------------------------------+
        | pc             | precipitation in inches               |
        +----------------+---------------------------------------+
        | temp           | air temperature in degrees fahrenheit |
        +----------------+---------------------------------------+
        | rhumid         | air relative humidity as percentage   |
        +----------------+---------------------------------------+
        | cndvty         | water electrical conductivity in      |
        |                | micromhos                             |
        +----------------+---------------------------------------+
        | tds            | total suspended solids                |
        +----------------+---------------------------------------+
        | windsp         | wind speed miles per hour             |
        +----------------+---------------------------------------+
        | winddir        | wind direction in degrees azimuth     |
        +----------------+---------------------------------------+
        | upperbasin     | ALL Upper Basin flow and water levels |
        +----------------+---------------------------------------+
        | lowerbasin     | ALL Lower Basin flow and water levels |
        +----------------+---------------------------------------+

    {start_date}
    {end_date}
    dam_site_location : str
        'head' (default) or 'tail'
        The site location relative to the dam.  Not used for `upperbasin`
        and `lowerbasin` parameters.
    """
    tsutils._printiso(lcra_hydromet(site_code,
                                    parameter_code,
                                    start_date=start_date,
                                    end_date=end_date,
                                    dam_site_location=dam_site_location))


def lcra_hydromet(site_code,
                  parameter_code,
                  start_date=None,
                  end_date=None,
                  dam_site_location='head'):
    r"""Hydrometeorologic data from the Lower Colorado River Authority."""
    from tsgettoolbox.services.lcra import hydromet

    df = hydromet.ulmo_df(site_code,
                          parameter_code,
                          start_date=start_date,
                          end_date=end_date,
                          dam_site_location=dam_site_location)

    return df


@mando.command('lcra_wq', formatter_class=HelpFormatter, doctype='numpy')
@tsutils.doc(tsutils.docstrings)
def lcra_wq_cli(site_code,
                historical=True,
                start_date=None,
                end_date=None):
    r"""Access data from the Lower Colorado River Authority Water Quality.

    Fetches historical or near real-time (for some sites) data.

    Lower Colorado River Authority: http://www.lcra.org

    Water Quality: http://waterquality.lcra.org/

    Parameters
    ----------
    site_code
        The site code to fetch data for. The following bay sites also have near
        real-time data available if `historical` option is set to False.

        +--------------------+-----------------+
        | Near Real-Time     | Name            |
        | (historical=False) |                 |
        | site_code          |                 |
        +====================+=================+
        | 6977               | Matagorda 4SSW  |
        +--------------------+-----------------+
        | 6985               | Matagorda 7 SW  |
        +--------------------+-----------------+
        | 6990               | Matagorda 8 SSW |
        +--------------------+-----------------+
        | 6996               | Matagorda 9 SW  |
        +--------------------+-----------------+

    historical
        Flag to indicate whether to get historical or near real-time data from
        the bay sites.
    {start_date}
    {end_date}"""
    tsutils._printiso(lcra_wq(site_code,
                              historical=historical,
                              start_date=start_date,
                              end_date=None))


def lcra_wq(site_code,
            historical=True,
            start_date=None,
            end_date=None):
    r"""Access data from the Lower Colorado River Authority Water Quality."""
    from tsgettoolbox.services.lcra import wq

    df = wq.ulmo_df(site_code=site_code,
                    historical=historical,
                    start_date=start_date,
                    end_date=end_date)

    return df




from __future__ import absolute_import

from tsgettoolbox.odo import convert
from tsgettoolbox.odo import odo
from tsgettoolbox.odo import resource

import pandas as pd

from tstoolbox import tsutils

# GHCND
# Global Historical Climatology Network - Daily

# pd.read_csv('ftp://ftp.ncdc.noaa.gov/pub/data/ghcn/daily/all/{0}.dly',
# skiprows=7,
# date_parser=testp,
# index_col=0,
# parse_dates=[[0,1]])


class ghcnd(object):
    def __init__(self, url, **query_params):
        params = {
            'station': None,
            'start_date': None,
            'end_date': None,
        }
        params.update(query_params)

        self.url = url
        self.query_params = params

# Function to make `resource` know about the new ghcnd type.


@resource.register(r'ftp://ftp.ncdc.noaa.gov/pub/data/ghcn/daily/all.*',
                   priority=17)
def resource_ghcnd(uri, **kwargs):
    return ghcnd(uri, **kwargs)

# Function to convert from ghcnd type to pd.DataFrame


@convert.register(pd.DataFrame, ghcnd)
def ghcnd_to_df(data, **kwargs):
    df = pd.read_fwf(data.url + '/' + data.query_params['station'] + '.dly',
                     widths=[11, 4, 2, 4] + [5, 1, 1, 1] * 31)
    newcols = ['station', 'year', 'month', 'code']
    for day in list(range(1, 32)):
        for col in ['', 'm', 'q', 's']:
            newcols.append('{0}{1:02}'.format(col, day))
    df.columns = newcols
    codes = ['TMAX',  # Temperature MAX (1/10 degree C)
             'TMIN',  # Temperature MIN (1/10 degree C)
             'PRCP',  # PReCiPitation (tenths of mm)
             'SNOW',  # SNOWfall (mm)
             'SNWD',  # SNoW Depth (mm)
             # Average cloudiness midnight to midnight from 30-second
             # ceilometer data (percent)
             'ACMC',
             # Average cloudiness midnight to midnight from manual observations
             # (percent)
             'ACMH',
             # Average cloudiness sunrise to sunset from 30-second ceilometer
             # data (percent)
             'ACSC',
             # Average cloudiness sunrise to sunset from manual observations
             # (percent)
             'ACSH',
             'AWDR',  # Average daily wind direction (degrees)
             'AWND',  # Average daily wind speed (tenths of meters per second)
             # Number of days included in the multiday evaporation total (MDEV)
             'DAEV',
             # Number of days included in the multiday precipiation total
             # (MDPR)
             'DAPR',
             # Number of days included in the multiday snowfall total (MDSF)
             'DASF',
             # Number of days included in the multiday minimum temperature
             # (MDTN)
             'DATN',
             # Number of days included in the multiday maximum temperature
             # (MDTX)
             'DATX',
             # Number of days included in the multiday wind movement (MDWM)
             'DAWM',
             # Number of days with non-zero precipitation included in multiday
             # precipitation total (MDPR)
             'DWPR',
             # Evaporation of water from evaporation pan (tenths of mm)
             'EVAP',
             # Time of fastest mile or fastest 1-minute wind (hours and
             # minutes, i.e., HHMM)
             'FMTM',
             'FRGB',  # Base of frozen ground layer (cm)
             'FRGT',  # Top of frozen ground layer (cm)
             'FRTH',  # Thickness of frozen ground layer (cm)
             'GAHT',  # Difference between river and gauge height (cm)
             # Multiday evaporation total (tenths of mm; use with DAEV)
             'MDEV',
             # Multiday precipitation total (tenths of mm; use with DAPR and
             # DWPR, if available)
             'MDPR',
             'MDSF',  # Multiday snowfall total
             # Multiday minimum temperature (tenths of degrees C; use with
             # DATN)
             'MDTN',
             # Multiday maximum temperature (tenths of degress C; use with
             # DATX)
             'MDTX',
             'MDWM',  # Multiday wind movement (km)
             # Daily minimum temperature of water in an evaporation pan (tenths
             # of degrees C)
             'MNPN',
             # Daily maximum temperature of water in an evaporation pan (tenths
             # of degrees C)
             'MXPN',
             'PGTM',  # Peak gust time (hours and minutes, i.e., HHMM)
             'PSUN',  # Daily percent of possible sunshine (percent)
             # Average temperature (tenths of degrees C) [Note that TAVG from
             # source 'S' corresponds to an average for the period ending at
             # 2400 UTC rather than local midnight]
             'TAVG',
             'THIC',  # Thickness of ice on water (tenths of mm)
             # Temperature at the time of observation (tenths of degrees C)
             'TOBS',
             'TSUN',  # Daily total sunshine (minutes)
             'WDF1',  # Direction of fastest 1-minute wind (degrees)
             'WDF2',  # Direction of fastest 2-minute wind (degrees)
             'WDF5',  # Direction of fastest 5-second wind (degrees)
             'WDFG',  # Direction of peak wind gust (degrees)
             'WDFI',  # Direction of highest instantaneous wind (degrees)
             'WDFM',  # Fastest mile wind direction (degrees)
             'WDMV',  # 24-hour wind movement (km)
             'WESD',  # Water equivalent of snow on the ground (tenths of mm)
             'WESF',  # Water equivalent of snowfall (tenths of mm)
             # Fastest 1-minute wind speed (tenths of meters per second)
             'WSF1',
             # Fastest 2-minute wind speed (tenths of meters per second)
             'WSF2',
             # Fastest 5-second wind speed (tenths of meters per second)
             'WSF5',
             'WSFG',  # Peak gust wind speed (tenths of meters per second)
             # Highest instantaneous wind speed (tenths of meters per second)
             'WSFI',
             'WSFM',  # Fastest mile wind speed (tenths of meters per second)
             ]

    # SN*# = Minimum soil temperature (tenths of degrees C)
    #        where * corresponds to a code
    #        for ground cover and # corresponds to a code for soil
    #        depth.
    #
    #        Ground cover codes include the following:
    #        0 = unknown
    #        1 = grass
    #        2 = fallow
    #        3 = bare ground
    #        4 = brome grass
    #        5 = sod
    #        6 = straw multch
    #        7 = grass muck
    #        8 = bare muck
    #
    #        Depth codes include the following:
    #        1 = 5 cm
    #        2 = 10 cm
    #        3 = 20 cm
    #        4 = 50 cm
    #        5 = 100 cm
    #        6 = 150 cm
    #        7 = 180 cm
    for i in range(9):
        for j in range(1, 8):
            codes.append('SN{0}{1}'.format(i, j))

    # SX*# = Maximum soil temperature (tenths of degrees C)
    #        where * corresponds to a code for ground cover
    #        and # corresponds to a code for soil depth.
    #        See SN*# for ground cover and depth codes.
    for i in range(9):
        for j in range(1, 8):
            codes.append('SX{0}{1}'.format(i, j))

    # WT** = Weather Type where ** has one of the following values:
    #
    #        01 = Fog, ice fog, or freezing fog (may include heavy fog)
    #        02 = Heavy fog or heaving freezing fog (not always distinquished
    #        from fog)
    #        03 = Thunder
    #        04 = Ice pellets, sleet, snow pellets, or small hail
    #        05 = Hail (may include small hail)
    #        06 = Glaze or rime
    #        07 = Dust, volcanic ash, blowing dust, blowing sand, or blowing
    #        obstruction
    #        08 = Smoke or haze
    #        09 = Blowing or drifting snow
    #        10 = Tornado, waterspout, or funnel cloud
    #        11 = High or damaging winds
    #        12 = Blowing spray
    #        13 = Mist
    #        14 = Drizzle
    #        15 = Freezing drizzle
    #        16 = Rain (may include freezing rain, drizzle, and freezing
    #        drizzle)
    #        17 = Freezing rain
    #        18 = Snow, snow pellets, snow grains, or ice crystals
    #        19 = Unknown source of precipitation
    #        21 = Ground fog
    #        22 = Ice fog or freezing fog
    codes.extend(['WT{0:02}'.format(i) for i in range(1, 23)])

    # WV** = Weather in the Vicinity where ** has one of the following values:
    #        01 = Fog, ice fog, or freezing fog (may include heavy fog)
    #        03 = Thunder
    #        07 = Ash, dust, sand, or other blowing obstruction
    #        18 = Snow or ice crystals
    #        20 = Rain or snow shower
    codes.extend(['WV{0:02}'.format(i) for i in [1, 3, 7, 18, 20]])

    for code in codes:
        tmpdf = df.ix[df['code'] == code, :]
        if len(tmpdf) == 0:
            continue
        tmpdf.set_index(['year', 'month'], inplace=True)
        tmpdf = tmpdf.ix[:, list(range(2, 126, 4))].stack()
        tmpdf.index = (tmpdf.index.get_level_values(0).astype(str).values
                       + '-'
                       + tmpdf.index.get_level_values(1).astype(str).values
                       + '-'
                       + tmpdf.index.get_level_values(2).astype(str).values)

        # Get rid of bad dates, for example April 31.
        tmpdf.index = pd.to_datetime(tmpdf.index, errors='coerce')
        tmpdf = tmpdf[pd.notnull(tmpdf.index)]

        tmpdf = pd.DataFrame(tmpdf)
        tmpdf.columns = [code]
        tmpdf = tmpdf.ix[tsutils.parsedate(data.query_params['start_date'],
                                           strftime='%Y-%m-%d'):
                         tsutils.parsedate(data.query_params['end_date'],
                                           strftime='%Y-%m-%d'), :]
        try:
            ndf = ndf.join(tmpdf)
        except NameError:
            ndf = tmpdf

    ndf.index.name = 'Datetime'
    ndf.replace(to_replace=[-9999], value=[None], inplace=True)
    return ndf


if __name__ == '__main__':
    r = resource(
        r'ftp://ftp.ncdc.noaa.gov/pub/data/ghcn/daily/all',
        station='ASN00075020',
        start_date='2000-01-01',
        end_date='2001-01-01',
    )

    as_df = odo(r, pd.DataFrame)
    print('ghcnd')
    print(as_df)

    r = resource(
        r'ftp://ftp.ncdc.noaa.gov/pub/data/ghcn/daily/all',
        station='ASN00075020',
        start_date='10 years ago',
        end_date='9 years ago',
    )

    as_df = odo(r, pd.DataFrame)
    print('ghcnd')
    print(as_df)



from __future__ import absolute_import, division, print_function

from collections import namedtuple, Iterator
from contextlib import contextmanager
from warnings import warn

from datashape import discover
import networkx as nx
import numpy as np
from toolz import concatv

from .compatibility import map, adjacency
from .utils import expand_tuples, ignoring


ooc_types = set()  # Out-of-Core types


class FailedConversionWarning(UserWarning):
    def __init__(self, src, dest, exc):
        self.src = src
        self.dest = dest
        self.exc = exc

    def __str__(self):
        return 'Failed on %s -> %s. Working around\nError message:\n%s' % (
            self.src.__name__, self.dest.__name__, self.exc,
        )


class IterProxy(object):
    """An proxy to another iterator to support swapping the underlying stream
    mid-iteration.

    Parameters
    ----------
    it : iterable
        The iterable to proxy.

    Attributes
    ----------
    it : iterable
        The iterable being proxied. This can be reassigned to change the
        underlying stream.
    """
    def __init__(self, it):
        self._it = iter(it)

    def __next__(self):
        return next(self.it)
    next = __next__  # py2 compat

    def __iter__(self):
        return self

    @property
    def it(self):
        return self._it

    @it.setter
    def it(self, value):
        self._it = iter(value)


class NetworkDispatcher(object):
    def __init__(self, name):
        self.name = name
        self.graph = nx.DiGraph()

    def register(self, a, b, cost=1.0):
        sigs = expand_tuples([a, b])

        def _(func):
            for a, b in sigs:
                self.graph.add_edge(b, a, cost=cost, func=func)
            return func
        return _

    def path(self, *args, **kwargs):
        return path(self.graph, *args, **kwargs)

    def __call__(self, *args, **kwargs):
        return _transform(self.graph, *args, **kwargs)


def _transform(graph, target, source, excluded_edges=None, ooc_types=ooc_types,
               **kwargs):
    """ Transform source to target type using graph of transformations """
    # take a copy so we can mutate without affecting the input
    excluded_edges = (excluded_edges.copy()
                      if excluded_edges is not None else
                      set())

    with ignoring(NotImplementedError):
        if 'dshape' not in kwargs or kwargs['dshape'] is None:
            kwargs['dshape'] = discover(source)

    pth = path(graph, type(source), target,
               excluded_edges=excluded_edges,
               ooc_types=ooc_types)

    x = source
    path_proxy = IterProxy(pth)
    for convert_from, convert_to, f, cost in path_proxy:
        try:
            x = f(x, excluded_edges=excluded_edges, **kwargs)
        except NotImplementedError as e:
            if kwargs.get('raise_on_errors'):
                raise
            warn(FailedConversionWarning(convert_from, convert_to, e))

            # exclude the broken edge
            excluded_edges |= {(convert_from, convert_to)}

            # compute the path from `source` to `target` excluding
            # the edge that broke
            fresh_path = list(path(graph, type(source), target,
                                   excluded_edges=excluded_edges,
                                   ooc_types=ooc_types))
            fresh_path_cost = path_cost(fresh_path)

            # compute the path from the current `convert_from` type
            # to the `target`
            try:
                greedy_path = list(path(graph, convert_from, target,
                                        excluded_edges=excluded_edges,
                                        ooc_types=ooc_types))
            except nx.exception.NetworkXNoPath:
                greedy_path_cost = np.inf
            else:
                greedy_path_cost = path_cost(greedy_path)

            if fresh_path_cost < greedy_path_cost:
                # it is faster to start over from `source` with a new path
                x = source
                pth = fresh_path
            else:
                # it is faster to work around our broken edge from our
                # current location
                pth = greedy_path

            path_proxy.it = pth

    return x


PathPart = namedtuple('PathPart', 'convert_from convert_to func cost')
_virtual_superclasses = (Iterator,)


def path(graph, source, target, excluded_edges=None, ooc_types=ooc_types):
    """ Path of functions between two types """
    if not isinstance(source, type):
        source = type(source)
    if not isinstance(target, type):
        target = type(target)

    for cls in concatv(source.mro(), _virtual_superclasses):
        if cls in graph:
            source = cls
            break

    # If both source and target are Out-Of-Core types then restrict ourselves
    # to the graph of out-of-core types
    if ooc_types:
        oocs = tuple(ooc_types)
        if issubclass(source, oocs) and issubclass(target, oocs):
            graph = graph.subgraph([n for n in graph.nodes()
                                    if issubclass(n, oocs)])
    with without_edges(graph, excluded_edges) as g:
        pth = nx.shortest_path(g, source=source, target=target, weight='cost')
        edge = adjacency(graph)

        def path_part(src, tgt):
            node = edge[src][tgt]
            return PathPart(src, tgt, node['func'], node['cost'])

        return map(path_part, pth, pth[1:])


def path_cost(path):
    """Calculate the total cost of a path.
    """
    return sum(p.cost for p in path)


@contextmanager
def without_edges(g, edges):
    edges = edges or []
    held = dict()
    _g_edge = adjacency(g)
    for a, b in edges:
        held[(a, b)] = _g_edge[a][b]
        g.remove_edge(a, b)

    try:
        yield g
    finally:
        for (a, b), kwargs in held.items():
            g.add_edge(a, b, **kwargs)
